/* Placed in the public domain 2000 by Sam Trenholme */
/* Parse a mararc file */

#include "../MaraDns.h"
#include "../libs/MaraHash.h"
#include "ParseCsv1_en.h"
#include <stdlib.h>
#include <pwd.h>
#include <sys/types.h>
#include <stdio.h>

/* Convert a dotted-decimal IP (in a js_string object) followed by
   a netmask in to a raw IP and netmask, using a ipv4pair object.
   The input data is in the form 10.69.69.69/24 or 10.69.69.69/255.255.0.0
   The output will be a binary IP and netmask, in network order.
   input: pointer to dotted decimal data, pointer to ipv4pair object to 
          place raw IP in to, offset from top to start looking
   output: JS_ERROR on error, pointer to first non-ip byte on SUCCESS
           (-2 if no non-ip byte was found)
*/

int ddip_ip_mask(js_string *ddip, ipv4pair *ips, int offset) {
    int qr, counter,firstoffset,slashp;
    int ret;

    unsigned char ip_byte;

    static js_string *dotq = 0;
    static js_string *numdotq = 0;
    static js_string *slashq = 0;
    static uint32 masks[33] = { 
    0x00000000, 
    0x80000000, 
    0xc0000000, 
    0xe0000000, 
    0xf0000000, 
    0xf8000000, 
    0xfc000000, 
    0xfe000000, 
    0xff000000, 
    0xff800000, 
    0xffc00000, 
    0xffe00000, 
    0xfff00000, 
    0xfff80000, 
    0xfffc0000, 
    0xfffe0000, 
    0xffff0000, 
    0xffff8000, 
    0xffffc000, 
    0xffffe000, 
    0xfffff000, 
    0xfffff800, 
    0xfffffc00, 
    0xfffffe00, 
    0xffffff00, 
    0xffffff80, 
    0xffffffc0, 
    0xffffffe0, 
    0xfffffff0, 
    0xfffffff8, 
    0xfffffffc, 
    0xfffffffe, 
    0xffffffff };

    char dot = '.';
    char slash = '/';

    /* Sanity checks */
    if(mara_goodjs(ddip) == JS_ERROR)
        return JS_ERROR;
    if(ips == 0)
        return JS_ERROR;

    /* Allocate string if this is the first time we are running this */
    if(dotq == 0) {
        if((dotq = js_create(7,1)) == 0)
	    return JS_ERROR;
        /* Place a dot in that string */
        if(js_str2js(dotq,&dot,1,1) == JS_ERROR) {
	    js_destroy(dotq);
	    dotq = 0;
            return JS_ERROR;
	    }
        }
    if(numdotq == 0) {
        if((numdotq = js_create(211,1)) == 0)
	    return JS_ERROR;
        if(js_set_encode(numdotq,JS_8859_1) == JS_ERROR) {
	    js_destroy(numdotq);
	    numdotq = 0;
	    return JS_ERROR;
	    }
        /* Place numbers in that string */
	if(js_numbers(numdotq) == JS_ERROR) {
	    js_destroy(numdotq);
	    numdotq = 0;
	    return JS_ERROR;
	    }
        /* Add the dot to that regex */
	if(js_append(dotq,numdotq) == JS_ERROR) {
	    js_destroy(numdotq);
	    numdotq = 0;
	    return JS_ERROR;
	    }
        }
    if(slashq == 0) {
        if((slashq = js_create(7,1)) == 0)
	    return JS_ERROR;
        /* Place a slash in that string */
        if(js_str2js(slashq,&slash,1,1) == JS_ERROR) {
	    js_destroy(slashq);
	    slashq = 0;
            return JS_ERROR;
	    }
        }

    /* Keep track of where we begin looking in the string */
    firstoffset = offset;

    /* See the first non number/dot after the offset */
    ret = js_notmatch_offset(numdotq,ddip,offset);
    /* If it returned an error, or the character at the offset is not
       a number nor a dot, return an error */
    if(ret == JS_ERROR || ret == offset)
        return JS_ERROR;

    /* Begin the ddip (dotted decimal IP) to raw binary ip conversion */
    if(js_length(ddip) < offset + 1)
        return JS_ERROR;

    ips->ip = 0;

    for(counter = 0; counter < 4; counter++) {
        if(offset == -2)
	    return JS_ERROR;

        qr = js_atoi(ddip,offset);

        if(qr < 0 || qr > 255)
            return JS_ERROR;

        ip_byte = qr;

        /* Add the byte in question to the ipv4pair object */
        ips->ip |= (ip_byte & 0xff) << (24 - 8 * counter); 

        /* Return error if there was a non-number/dot in the IP */
        if(counter == 4 && ret < offset)
            return JS_ERROR;

        offset = js_match_offset(dotq,ddip,offset + 1);
	if(offset == JS_ERROR)
	    return JS_ERROR;

        offset++;
        }

    /* If no slash immediately after, assumes the netmask is 255.255.255.255 */
    if(js_match_offset(slashq,ddip,firstoffset) != ret ||
       js_match_offset(slashq,ddip,firstoffset) == -2) {
        ips->mask = 0xffffffff;
        return ret;
	}
    slashp = ret;

    /* There are two possible formats we have to deal with:
       10.69.69.69/255.255.255.0 and 10.69.69.69/24.  */
    ret = js_match_offset(dotq,ddip,slashp + 1);
    if(ret == JS_ERROR) 
        return JS_ERROR;
    offset = js_notmatch_offset(numdotq,ddip,slashp + 1);
    if(offset == JS_ERROR)
        return JS_ERROR;
    /* If it is in a 10.69.69.69/24 form */
    if(ret == -2 || (offset < ret && offset != -2)) {
        ret = js_atoi(ddip,slashp + 1);
        if(ret < 0 || ret > 32)
            return JS_ERROR;
        ips->mask = masks[ret];
	if(offset + 1 < ddip->unit_count)
            return offset + 1;
        else
	    return -2;
	}
    else {
        offset = slashp + 1;
	ips->mask = 0;
        for(counter = 0; counter < 4; counter++) {
            if(offset == -2)
	        return JS_ERROR;

            qr = js_atoi(ddip,offset);

            if(qr < 0 || qr > 255)
                return JS_ERROR;

            ip_byte = qr;

            /* Add the byte in question to the ipv4pair object */
            ips->mask |= (ip_byte & 0xff) << (24 - 8 * counter); 

            if(counter != 3) {
                offset = js_match_offset(dotq,ddip,offset + 1);
	        if(offset == JS_ERROR)
	            return JS_ERROR;
                }
            else
                return js_notmatch_offset(numdotq,ddip,offset + 1);

            offset++;
            }
        return JS_ERROR; /* We should never get here */
        }
    
    return JS_ERROR; /* We should never get here */
    }

