/* Placed in the public domain 2001 by Sam Trenholme */

/* This is the core DNS server */

/* Language specifi labels */
#include "MaraDNS_locale.h"

/* Include stuff needed to be a UDP server */

#include "../libs/MaraHash.h"
#include "../MaraDns.h"
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#ifdef __FreeBSD__
#include <sys/time.h>
#endif
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "../dns/functions_dns.h"
#include "../parse/functions_parse.h"
#include "functions_server.h"


/* Our two global variables: A pointer to the "big hash", log reporting
   level */
mhash *bighash;
int log_level = 0; /* 0: No messages ever, 1: Startup and shutdown messages,
                  2: Log queries that generate errors, 3: Log all queries */ 
int no_fingerprint = 0; /* 0: Have some MaraDNS-specific features, such as
                              DDIP expansion and a special query that
                              tells you the version number of the server.
                           1: Attempts to have as generic an interface as
                              possible. */

/* Signal handler for SIGPIPE, so we don't terminate */
void handle_sigpipe() {
    printf("%s%s",L_CAUGHT_PIPE,L_N); /* "Caught SIGPIPE" */
    return;
    }

/* Signal handler for other signals */
void handle_signal() {
    printf("%s%s",L_CAUGHT_SIG,L_N); /* "Caught Signal" */
    return;
    }

/* Print out log messages
   Input: Null-terminated string with the message to log
   Output: JS_SUCCESS on success, JS_ERROR on error
*/

int mlog(char *logmessage) {

    if(log_level == 0)
        return JS_SUCCESS;

    if(logmessage == 0)
        return JS_ERROR;
    printf("%s%s%s",L_LOG,logmessage,L_N); /* "Log: ", logmessage, "\n" */

    return JS_SUCCESS;
    }

/* Print out log messages of js_string messages
   Input: js_string object to log
   Output: JS_SUCCESS on success, JS_ERROR on error
*/

int jlog(js_string *logmessage) {

    int ret;

    if(log_level == 0)
        return JS_SUCCESS;

    printf("%s",L_LOG); /* "Log: " */
    ret = show_esc_stdout(logmessage);
    printf("%s",L_N); /* "\n" */
    return ret;
    }

/* Print out log message of a Null-terminated string followed by a js_string
   Input: Null-terminated string, js_string
   Output: JS_SUCCESS on success, JS_ERROR on error */

int zjlog(char *left, js_string *right) {
    int ret;
    if(log_level == 0)
        return JS_SUCCESS;
    if(left == 0)
        return JS_ERROR;
    printf("%s%s",L_LOG,left); /* "Log: ", left */
    ret = show_esc_stdout(right);
    printf("%s",L_N); /* "\n" */
    return ret;
    }

/* Handler to handle fatal errors.
   Input: Pointer to null-terminalted string with fatal error
   Output: MaraDNS exits
*/

void harderror(char *why) {
    printf("%s%s%s",L_FATAL,why,L_N); /* "Fatal Error: ", why, "\n" */
    exit(3);
    }

/* Handler to handle system fatal errors.
   Input: Pointer to null-terminalted string with fatal error
   Output: MaraDNS exits
*/
 
void sys_harderror(char *why) {
    printf("%s%s%s",L_FATAL,why,L_N); /* "Fatal Error: ", why, "\n" */
    perror(L_SYSERROR); /* "System said: " */
    exit(3);
    }

/* Return a packet indicating that there was an error in the received
   packet
   input: socket number,
          a js_string object that we get the data from the first two
          bytes from, a sockaddr of who to send the error to, the error 
	  to give them in the RCOADE part of the header
   output: JS_ERROR on error, JS_SUCCESS on success
*/

int udperror(int sock,js_string *raw, struct sockaddr *from, int error) {

    q_header header;
    js_string *reply;
    int len_inet = sizeof(struct sockaddr);

    if(log_level >= 2)
        zjlog(L_BAD_QUERY,raw); /* "Bad query received: " */

    if(raw->unit_count < 2 || raw->max_count < 3)
        return JS_ERROR;
  
    if((reply = js_create(32,1)) == 0)
        return JS_ERROR;

    /* Fill out the header */
    header.id = ((*(raw->string) & 0x0f) << 8) | (*(raw->string + 1) & 0x0f);
    header.qr = 1;
    header.opcode = 0;
    header.aa = 0; /* Errors are never authoritative (unless they are 
                      NXDOMAINS, which this is not) */
    header.tc = 0;
    header.rd = 0;
    header.ra = 0;
    header.z = 0;
    header.rcode = error;
    header.qdcount = 0;
    header.ancount = 0;
    header.nscount = 0;
    header.arcount = 0;

    /* Make that raw UDP data */
    if(make_hdr(&header,reply) == JS_ERROR)
        return JS_ERROR;
   
    /* Send them the reply */
    sendto(sock,reply->string,reply->unit_count,0,from,len_inet);

    return JS_SUCCESS;

    }

/* If we successfully found a record, spit out that record on the 
   udp packet.
   Input: Where a pointer to the rr in question is, the id of the
          query they sent us, the socket the
          UDP bind is on, the sockaddr of the client who sent us the message,
          a js_string containing the query (dname + type)
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int udpsuccess(rr *where,int id,int sock,struct sockaddr *client,
               js_string *query) {
    js_string *most, *ar; /* Most of the data then the additional records */
   
    uint16 first_rr_type;
    int in_ns = 0; 
    int length_save;
    int len_inet = sizeof(struct sockaddr);

    q_header header;
    /* Initialize the js_string objects */
    if((most = js_create(1024,1)) == 0)
        return JS_ERROR;
    if((ar = js_create(1024,1)) == 0) {
        js_destroy(most);
        return JS_ERROR;
        }

    /* Make the header a placeholder for now */
    header.id = id;
    if(make_hdr(&header,most) == JS_ERROR)
        goto giveerror;

    /* Sanity check */
    if(where == 0) {
	goto giveerror;
	}
    if(where->query == 0) {
	goto giveerror;
	}
    if(js_has_sanity(where->query) == JS_ERROR) {
	goto giveerror;
	}
    if(where->data == 0) {
	goto giveerror;
	}
    if(js_has_sanity(where->data) == JS_ERROR) {
	goto giveerror;
	}
    if(js_has_sanity(query) == JS_ERROR) {
	goto giveerror;
	}
    first_rr_type = get_rtype(query);

    /* We have to add this header here--authoritative depends on the 
       authorative status of the first record we find */
    header.aa = where->authoritative;

    /* The data must be between 0 and 65535 bytes in length (16-bit
       unsigned value) */
    if(where->data->unit_count < 0 || where->data->unit_count > 65535) {
	goto giveerror;
	}
   
    /* Append the question to the answer */
    if(js_append(query,most) == JS_ERROR) {
	goto giveerror;
	}

    /* Append the class (in) to the answer */
    if(js_adduint16(most,1) == JS_ERROR) {
	goto giveerror;
	}

    /* We will increment the ancount, nscount, and arcount, starting at 0 */
    header.ancount = 0;
    header.nscount = 0;
    header.arcount = 0;

    /* OK, we now add the answers */
    while(where != 0) {
        /* Increment the number of answers -or- ns records */
        if(first_rr_type != RR_NS && where->rr_type == RR_NS)
	    in_ns = 1;
        if(!in_ns) 
            header.ancount++;
        else 
            header.nscount++;
        /* Append the name for this answer to the answer */
        if(js_append(where->query,most) == JS_ERROR) {
	    goto giveerror;
	    }
        /* Append the class (in) to the answer */
        if(js_adduint16(most,1) == JS_ERROR) {
	    goto giveerror;
	    }
        /* Append the ttl to the answer */
        if(js_adduint32(most,where->ttl) == JS_ERROR) {
	    goto giveerror;
	    }
        /* Add the rdlength to the answer */
        if(js_adduint16(most,where->data->unit_count) == JS_ERROR) {
	    goto giveerror;
	    }
        /* Add the record itself to the answer */
        if(js_append(where->data,most) == JS_ERROR) {
	    goto giveerror;
	    }
        /* If there is an IP, and this is *not* a CNAME record,
           append the IP of the answer to the AR section */
        if(where->ip != 0 && where->rr_type != RR_CNAME) {
	    /* Increment the number of additional records */
	    header.arcount++;
	    /* Append the name for this answer to the ip */
	    if(js_append(where->ip->query,ar) == JS_ERROR) {
	        goto giveerror;
		}
            /* Append the class (in) to the ip */
            if(js_adduint16(ar,1) == JS_ERROR) {
	        goto giveerror;
		}
            /* Append the TTL to the ip */
	    if(js_adduint32(ar,where->ip->ttl) == JS_ERROR) {
	        goto giveerror;
		}
            /* Add the rdlength to the ip */
	    if(js_adduint16(ar,where->ip->data->unit_count) == JS_ERROR) {
	        goto giveerror;
		}
            /* Add the record itself to the ip */
	    if(js_append(where->ip->data,ar) == JS_ERROR) {
	        goto giveerror;
		}
            }
        /* If this is an IP, and this is a CNAME record,
	   append the IP in question to the answer section */
        if(where->ip != 0 && where->rr_type == RR_CNAME) {
	    /* Increment the number of answer records */
	    header.ancount++;
            /* Append the name for this answer to the ip */
	    if(js_append(where->ip->query,most) == JS_ERROR)
	        goto giveerror;
            /* Append the class (in) to the ip */
	    if(js_adduint16(most,1) == JS_ERROR)
	        goto giveerror;
            /* Append the TTL to the ip */
            if(js_adduint32(most,where->ip->ttl) == JS_ERROR)
                goto giveerror;
            /* Add the rdlength to the ip */
	    if(js_adduint16(most,where->ip->data->unit_count) == JS_ERROR)
	        goto giveerror;
            /* Add the record itself to the ip */
	    if(js_append(where->ip->data,most) == JS_ERROR)
	        goto giveerror;
            }
        /* Go on to the next record in the linked list */
        where = where->next;
        }

    /* Customize the header */
    /* header.id already set */
    header.qr = 1;
    header.opcode = 0;
    header.tc = 0; 
    header.rd = 0;
    header.ra = 0;
    header.z = 0;
    header.rcode = 0; /* No error */
    header.qdcount = 1;

    /* OBhack: Tack on the header at the beginning without molesting the
       rest of the string */
    length_save = most->unit_count;
    make_hdr(&header,most);
    most->unit_count = length_save;

    /* Add the ar records to the end */
    if(js_append(ar,most) == JS_ERROR) {
        goto giveerror; 
	}

    /* Compress "most" and place the compressed data in "ar" */
    if(compress_data(most,ar) == JS_ERROR) {
        js_destroy(ar); 
        udperror(sock,most,client,SERVER_FAIL); 
        js_destroy(most);
        return JS_ERROR;
        }

    /* Check to make sure the data fits in under 512 bytes */
    if(ar->unit_count > 512) {
        /* We handle truncation by truncating everything except the
	   12-byte header */
        header.tc = 1;
        make_hdr(&header,ar);
	}

    /* Success! Put out the good data */
    sendto(sock,ar->string,ar->unit_count,0,client,len_inet);

    js_destroy(most);
    js_destroy(ar);

    return JS_SUCCESS;

    /* We use gotos to make up for C's lack of error trapping */
    giveerror:
        js_destroy(ar);
        udperror(sock,most,client,SERVER_FAIL); 
	js_destroy(most);
	return JS_ERROR;

    }

/* If we successfully found a record, add the answer to the A record,
   if applicable, add the NS data too, and add the appropriate 
   additional records.
   Input: Where a pointer to the rr in question is, a pointer to the
          string where we add answers, pointer to ns data string, a pointer 
          to the string where we add additional records, a pointer to the
          number containing the number of answers, a pointer to the
          number containing the number of authority records, a pointer to
          the number containing the number of additional records, whether
          to add records to the authority (ns) section.
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int add_answer(rr *where,js_string *most, js_string *ns, js_string *ar, 
               uint16 *ancount, uint16 *nscount, uint16 *arcount, 
               int add_ns) {
   
    uint16 first_rr_type;
    int in_ns = 0; 

    /* Sanity check */
    if(where == 0) {
	goto giveerror;
	}
    if(where->query == 0) {
	goto giveerror;
	}
    if(js_has_sanity(where->query) == JS_ERROR) {
	goto giveerror;
	}
    if(where->data == 0) {
	goto giveerror;
	}
    if(js_has_sanity(where->data) == JS_ERROR) {
	goto giveerror;
	}
    first_rr_type = get_rtype(where->query);

    /* The data must be between 0 and 65535 bytes in length (16-bit
       unsigned value) */
    if(where->data->unit_count < 0 || where->data->unit_count > 65535) {
	goto giveerror;
	}
   
    /* OK, we now add the answers */
    while(where != 0) {
        /* Increment the number of answers -or- ns records */
        if(first_rr_type != RR_NS && where->rr_type == RR_NS) {
            if(add_ns == 1) 
	        in_ns = 1;
            else 
                return JS_SUCCESS;
            }
        if(!in_ns) {
            (*ancount)++;
            /* Append the name for this answer to the answer */
            if(js_append(where->query,most) == JS_ERROR) {
	        goto giveerror;
	        }
            /* Append the class (in) to the answer */
            if(js_adduint16(most,1) == JS_ERROR) {
	        goto giveerror;
	        }
            /* Append the ttl to the answer */
            if(js_adduint32(most,where->ttl) == JS_ERROR) {
	        goto giveerror;
	        }
            /* Add the rdlength to the answer */
            if(js_adduint16(most,where->data->unit_count) == JS_ERROR) {
	        goto giveerror;
	        }
            /* Add the record itself to the answer */
            if(js_append(where->data,most) == JS_ERROR) {
	        goto giveerror;
	        }
            }
        else {
            /* Append the name for this answer to the answer */
            if(js_append(where->query,ns) == JS_ERROR) {
	        goto giveerror;
	        }
            /* Append the class (in) to the answer */
            if(js_adduint16(ns,1) == JS_ERROR) {
	        goto giveerror;
	        }
            /* Append the ttl to the answer */
            if(js_adduint32(ns,where->ttl) == JS_ERROR) {
	        goto giveerror;
	        }
            /* Add the rdlength to the answer */
            if(js_adduint16(ns,where->data->unit_count) == JS_ERROR) {
	        goto giveerror;
	        }
            /* Add the record itself to the answer */
            if(js_append(where->data,ns) == JS_ERROR) {
	        goto giveerror;
	        }
            (*nscount)++;
            }
        /* If there is an IP, and this is *not* a CNAME record,
           append the IP of the answer to the AR section */
        if(where->ip != 0 && where->rr_type != RR_CNAME) {
	    /* Increment the number of additional records */
	    (*arcount)++;
	    /* Append the name for this answer to the ip */
	    if(js_append(where->ip->query,ar) == JS_ERROR) {
	        goto giveerror;
		}
            /* Append the class (in) to the ip */
            if(js_adduint16(ar,1) == JS_ERROR) {
	        goto giveerror;
		}
            /* Append the TTL to the ip */
	    if(js_adduint32(ar,where->ip->ttl) == JS_ERROR) {
	        goto giveerror;
		}
            /* Add the rdlength to the ip */
	    if(js_adduint16(ar,where->ip->data->unit_count) == JS_ERROR) {
	        goto giveerror;
		}
            /* Add the record itself to the ip */
	    if(js_append(where->ip->data,ar) == JS_ERROR) {
	        goto giveerror;
		}
            }
        /* To do: A records attached to CNAMES are added as a second AN
                  record if the originally requested query was not a CNAME
        */
        /* Go on to the next record in the linked list */
        where = where->next;
        }

    return JS_SUCCESS;

    /* We use gotos to make up for C's lack of error trapping */
    giveerror:
	return JS_ERROR;

    }

/* If they asked for a RR_ANY record, do this: 
   1. See if MX records exist.  If so, add it to the answer to give.
   2. See if A records record exist.  If so, add it.
   3. If neither A nor MX exist, look for a CNAME record.  If so,
      return just the CNAME record.
   4. Otherwise, return "query denied".
   Input: ID of the iquery they sent us, socket of the request, a sockaddr
          with their address and port number on it, a js_string containing
          the query (dname + type)
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int udpany(int id,int sock,struct sockaddr *client, js_string *query) {
    js_string *most, *ns, *ar; /* The answers, the ns records, the ar records*/
   
    int length_save;
    int len_inet = sizeof(struct sockaddr);
    int found = 0;
    mhash_e spot_data;

    q_header header;
    /* Initialize the js_string objects */
    if((most = js_create(1024,1)) == 0)
        return JS_ERROR;
    if((ar = js_create(1024,1)) == 0) {
        js_destroy(most);
        return JS_ERROR;
        }
    if((ns = js_create(1024,1)) == 0) {
        js_destroy(most); js_destroy(ar);
        return JS_ERROR;
        }

    /* Make the header a placeholder for now */
    header.id = id;
    if(make_hdr(&header,most) == JS_ERROR)
        goto giveerror;

    /* Append the question to the answer */
    if(js_append(query,most) == JS_ERROR) {
	goto giveerror;
	}

    /* Append the class (in) to the answer */
    if(js_adduint16(most,1) == JS_ERROR) {
	goto giveerror;
	}

    /* We will increment the ancount, nscount, an arcount, starting at 0 */
    header.ancount = 0;
    header.nscount = 0;
    header.arcount = 0;

    /* Start synthesizing the reply */

    /* Look for an A record with the same name as the query */
    if(change_rtype(query,RR_A) == JS_ERROR)
        goto giveerror;
    spot_data = mhash_get(bighash,query);
    /* If found, add the data to our records */
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
        found = 1;
        if(add_answer(spot_data.value,most,ns,ar,&(header.ancount),
	           &(header.nscount),&(header.arcount),1) == JS_ERROR)
            goto giveerror;
        }
    /* Look for MX record with the same name as the query */
    if(change_rtype(query,RR_MX) == JS_ERROR)
        goto giveerror;
    spot_data = mhash_get(bighash,query);
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
        if(found == 1) {
	    if(add_answer(spot_data.value,most,ns,ar,&(header.ancount),
	       &(header.nscount),&(header.arcount),0) == JS_ERROR)
	        goto giveerror;
            }
        else {
	    if(add_answer(spot_data.value,most,ns,ar,&(header.ancount),
	       &(header.nscount),&(header.arcount),1) == JS_ERROR)
	        goto giveerror;
            }
        found = 1;
	}

    /* If not found, look for lower-case version of the same query */
    if(found != 1) {
        found = fold_case(query);
        if(found == JS_ERROR)
	    goto giveerror;
        if(found == 1) /* Case folded */ {
	    found = 0;
	    /* Look for lower case version of A record */
	    if(change_rtype(query,RR_A) == JS_ERROR) 
	        goto giveerror;
            spot_data = mhash_get(bighash,query);
	    /* If A record of lower-case found... */
	    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
	        found = 1;
		if(add_answer(spot_data.value,most,ns,ar,&(header.ancount),
		   &(header.nscount),&(header.arcount),1) == JS_ERROR)
		    goto giveerror;
	        } 
            if(change_rtype(query,RR_MX) == JS_ERROR)
	        goto giveerror;
            spot_data = mhash_get(bighash,query);
	    /* If MX record of lower-case found... */
	    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
	        if(found == 1) {
		    if(add_answer(spot_data.value,most,ns,ar,&(header.ancount),
		       &(header.nscount),&(header.arcount),0) == JS_ERROR)
		        goto giveerror;
                    }
                else {
		    if(add_answer(spot_data.value,most,ns,ar,&(header.ancount),
		       &(header.nscount),&(header.arcount),1) == JS_ERROR)
		        goto giveerror;
                    }
                }
	    }
	}

    /* If nothing found, look for a CNAME record with the same name as 
       the query */
    if(found == 0) {
        if(change_rtype(query,RR_CNAME) == JS_ERROR)
            goto giveerror;
        spot_data = mhash_get(bighash,query);
        /* If found, add the data to our records */
        if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
            found = 1;
	    if(add_answer(spot_data.value,most,ns,ar,&(header.ancount),
	                 &(header.nscount),&(header.arcount),1) == JS_ERROR)
                goto giveerror;
            }
        }

    /* If nothing found, look for DDIP notation */
    if(found == 0) {
        if(change_rtype(query,RR_ANY) == JS_ERROR)
	    goto giveerror;
        found = ddip_check(id,sock,client,query);
	if(found == JS_ERROR)
	    goto giveerror;
        if(found == JS_SUCCESS)
            return JS_SUCCESS;
        }

    /* Return query refused if no answer was found */
    if(header.ancount == 0) {
        js_destroy(ar); 
        udperror(sock,most,client,REFUSED);
        js_destroy(most);
        return JS_SUCCESS;
	}

    /* Customize the header */
    /* header.id already set */
    header.qr = 1;
    header.opcode = 0;
    header.tc = 0; 
    header.rd = 0;
    header.ra = 0;
    header.aa = 0; /* XXX fix me */
    header.z = 0;
    header.rcode = 0; /* No error */
    header.qdcount = 1;

    /* OBhack: Tack on the header at the beginning without molesting the
       rest of the string */
    length_save = most->unit_count;
    make_hdr(&header,most);
    most->unit_count = length_save;

    /* Add the ns and ar records to the end */
    if(js_append(ns,most) == JS_ERROR) {
        goto giveerror; 
	}
    if(js_append(ar,most) == JS_ERROR) {
        goto giveerror; 
	}

    /* Compress "most" and place the compressed data in "ar" */
    if(compress_data(most,ar) == JS_ERROR) {
        js_destroy(ar); 
        udperror(sock,most,client,SERVER_FAIL); 
        js_destroy(most);
        return JS_ERROR;
        }

    /* Check to make sure the data fits in under 512 bytes */
    if(ar->unit_count > 512) {
        /* We handle truncation by truncating everything except the
	   12-byte header */
        header.tc = 1;
        make_hdr(&header,ar);
	}

    /* Success! Put out the good data */
    sendto(sock,ar->string,ar->unit_count,0,client,len_inet);

    js_destroy(most);
    js_destroy(ar);

    return JS_SUCCESS;

    /* We use gotos to make up for C's lack of error trapping */
    giveerror:
        js_destroy(ar);
        js_destroy(ns);
        udperror(sock,most,client,SERVER_FAIL); 
	js_destroy(most);
	return JS_ERROR;

    }

/* OK, there are a handful of record types which MaraDNS gives special
   treatment to when a TXT record is asked for the host name in question.
   This routine handles these special domain names.
   Input: ID of the query they sent us, socket of the request, a sockaddr
          with their address and port on it, a js_string containing
          the query (dname + type), The host name that is given special
          treatment (in a pre-hname2rfc1035 format), query type to convert,
          2 strings whose data is dependent on the the query_type to 
          convert.
*/

int easter_egg(int id,int sock,struct sockaddr *client, js_string *query,
               char *hname, uint16 type, char *opt1, char *opt2) {
    js_string *reply, *hname_js, *data; /* The reply, the query, the answer */
    q_header header;
    int result;

    /* Sanity checks */
    if(js_has_sanity(query) == JS_ERROR)
        return JS_ERROR;
    if(hname == 0 || opt1 == 0)
        return JS_ERROR;

    if((reply = js_create(512,1)) == 0)
        return JS_ERROR;
    if((hname_js = js_create(256,1)) == 0) {
        js_destroy(reply);
        return JS_SUCCESS;
        } 
    if((data = js_create(256,1)) == 0) {
        js_destroy(reply); js_destroy(hname_js);
        return JS_SUCCESS;
        } 

    /* Make sure that this is the query that they asked for */
    hname_js->encoding = query->encoding;

    if(js_qstr2js(hname_js,hname) == JS_ERROR)
        goto cleanup;

    if(hname_2rfc1035(hname_js) <= 0)
        goto cleanup;

    if(js_adduint16(hname_js,type) == JS_ERROR)
        goto cleanup;

    result = js_issame(hname_js,query);
    if(result == JS_ERROR)
        goto cleanup;

    if(result != 1) {
        js_destroy(reply); js_destroy(hname_js); js_destroy(data);
	return 0;
	}

    /* OK, the hostname matches the "easter egg" name, now we form
       the "easter egg" reply */
    
    /* Get the data from the options */
    /* If we ever support easter eggs for anything besides TXT
       records, this will become a switch statement */
    if(type != RR_TXT) {
        js_destroy(reply); js_destroy(hname_js); js_destroy(data);
	return 0;
	}

    if(opt2 == 0)
        return JS_ERROR;

    /* With TXT records, we take the string in opt1, add the string in
       opt2 to the string, and make that the data.  hname_js is used
       as a "throwaway" string */
    if(js_qstr2js(hname_js,"") == JS_ERROR)
        goto cleanup;
    if(js_qappend(opt1,hname_js) == JS_ERROR)
        goto cleanup;
    if(js_qappend(opt2,hname_js) == JS_ERROR)
        goto cleanup;
    if(js_qstr2js(data,"") == JS_ERROR)
        goto cleanup;
    if(hname_js->unit_count > 255)
        goto cleanup;
    if(js_addbyte(data,hname_js->unit_count) == JS_ERROR)
        goto cleanup;
    if(js_append(hname_js,data) == JS_ERROR)
        goto cleanup;

    /* Build up the header for this reply */
    if(id > 0 && id < 65535)
        header.id = id;
    else 
        goto cleanup;

    header.qr = 1; /* Reply */
    header.opcode = 0; /* Normal DNS */
    header.aa = 0; /* DDIP to A translations are never authoritative */
    header.tc = 0; /* A labels are too short to be truncated */
    header.rd = 0; /* Recursion not desired */
    header.ra = 0; /* Recursion not available */
    header.z = 0; /* This must be 0 unless we are EDNS aware (we aren't) */
    header.rcode = 0; /* Success! */
    header.qdcount = 1;
    header.ancount = 1;
    header.nscount = 0;
    header.arcount = 0;

    /* Make a header of the reply */
    if(make_hdr(&header,reply) == JS_ERROR) 
        goto cleanup;

    /* Add the question they asked to the reply */
    if(js_append(query,reply) == JS_ERROR) 
        goto cleanup;

    /* Add the class (in) to the answer */
    if(js_adduint16(reply,1) == JS_ERROR) 
        goto cleanup;

    /* We will now add out manufactured reply */
    if(js_append(query,reply) == JS_ERROR) 
        goto cleanup;
    /* Append the class (in) to the answer */
    if(js_adduint16(reply,1) == JS_ERROR) 
        goto cleanup;
    /* Append a bogus TTL to the answer */
    if(js_adduint32(reply,730110) == JS_ERROR) 
        goto cleanup;
    /* Add the rdlength to the answer */
    if(js_adduint16(reply,data->unit_count) == JS_ERROR) 
        goto cleanup;
    /* Add the actual data to the answer */
    if(js_append(data,reply) == JS_ERROR) 
        goto cleanup;

    /* Send the reply out */
    sendto(sock,reply->string,reply->unit_count,0,client,
           sizeof(struct sockaddr));
   
    /* And, we are done */
    js_destroy(reply);
    js_destroy(hname_js);
    js_destroy(data);
    return JS_SUCCESS;

    /* We use gotos to work around C's lack of error trapping */
    cleanup:
        js_destroy(reply);
        js_destroy(hname_js);
        js_destroy(data);
	return JS_ERROR;

    }

/* If we successfully found a star record, spit out that record on the 
   udp packet.
   Input: Where a pointer to the rr in question is, the id of the
          query they sent us, the socket the
          UDP bind is on, the sockaddr of the client who sent us the message,
          a js_string containing the query (dname + type), 
	  a js_string containing the answer.
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int udpstar(rr *where,int id,int sock,struct sockaddr *client,
               js_string *query, js_string *answer) {
    js_string *most, *ar; /* Most of the data then the additional records */
   
    uint16 first_rr_type;
    int in_ns = 0; 
    int length_save;
    int len_inet = sizeof(struct sockaddr);

    q_header header;
    /* Initialize the js_string objects */
    if((most = js_create(1024,1)) == 0)
        return JS_ERROR;
    if((ar = js_create(1024,1)) == 0) {
        js_destroy(most);
        return JS_SUCCESS;
        }

    /* Make the header a placeholder for now */
    header.id = id;
    if(make_hdr(&header,most) == JS_ERROR)
        goto giveerror;

    /* Sanity check */
    if(where == 0) {
        goto giveerror; 
	}
    if(where->query == 0) {
        goto giveerror; 
	}
    if(js_has_sanity(where->query) == JS_ERROR) {
        goto giveerror; 
	}
    if(where->data == 0) {
        goto giveerror;
	}
    if(js_has_sanity(where->data) == JS_ERROR) {
        goto giveerror;
	}
    if(js_has_sanity(query) == JS_ERROR) {
        goto giveerror;
	}
    first_rr_type = get_rtype(query);

    /* We have to add this header here--authoritative depends on the 
       authorative status of the first record we find */
    header.aa = where->authoritative;

    /* The data must be between 0 and 65535 bytes in length (16-bit
       unsigned value) */
    if(where->data->unit_count < 0 || where->data->unit_count > 65535) {
        goto giveerror;
	}
   
    /* Append the question to the answer */
    if(js_append(query,most) == JS_ERROR) {
        goto giveerror; 
	}

    /* Append the class (in) to the answer */
    if(js_adduint16(most,1) == JS_ERROR) {
        goto giveerror; 
	}

    /* We will increment the ancount, nscount, an arcount, starting at 0 */
    header.ancount = 0;
    header.nscount = 0;
    header.arcount = 0;

    /* OK, we now add the answers */
    while(where != 0) {
        /* Increment the number of answers -or- ns records */
        if(first_rr_type != RR_NS && where->rr_type == RR_NS)
	    in_ns = 1;
        if(!in_ns) 
            header.ancount++;
        else 
            header.nscount++;
        /* Append the name for the user's "canonical" query to the answer */
	if(!in_ns) {
            if(js_append(answer,most) == JS_ERROR) {
                goto giveerror; 
		}
            }
        /* (Unless we are telling them the NS records for this RR) */
        else {
	    if(js_append(where->query,most) == JS_ERROR) {
                goto giveerror; 
		}
            }
        /* Append the class (in) to the answer */
        if(js_adduint16(most,1) == JS_ERROR) {
            goto giveerror;
	    }
        /* Append the ttl to the answer */
        if(js_adduint32(most,where->ttl) == JS_ERROR) {
            goto giveerror;
	    }
        /* Add the rdlength to the answer */
        if(js_adduint16(most,where->data->unit_count) == JS_ERROR) {
            goto giveerror; 
	    }
        /* Add the record itself to the answer */
        if(js_append(where->data,most) == JS_ERROR) {
            goto giveerror;
	    }
        /* If there is an IP, and this is *not* a CNAME record,
           append the IP of the answer to the AR section */
        if(where->ip != 0 && where->rr_type != RR_CNAME) {
	    /* Increment the number of additional records */
	    header.arcount++;
	    /* Append the name for this answer to the ip */
	    if(js_append(where->ip->query,ar) == JS_ERROR) {
	        goto giveerror;
		}
            /* Append the class (in) to the ip */
            if(js_adduint16(ar,1) == JS_ERROR) {
	        goto giveerror;
		}
            /* Append the TTL to the ip */
	    if(js_adduint32(ar,where->ip->ttl) == JS_ERROR) {
	        goto giveerror;
		}
            /* Add the rdlength to the ip */
	    if(js_adduint16(ar,where->ip->data->unit_count) == JS_ERROR) {
	        goto giveerror;
		}
            /* Add the record itself to the ip */
	    if(js_append(where->ip->data,ar) == JS_ERROR) {
	        goto giveerror;
		}
            }
        /* To do: A records attached to CNAMES are added as a second AN
                  record if the originally requested query was not a CNAME
        */
        /* Go on to the next record in the linked list */
        where = where->next;
        }

    /* Customize the header */
    /* header.id already set */
    header.qr = 1;
    header.opcode = 0;
    header.tc = 0; /* To do: truncation handling */
    header.rd = 0;
    header.ra = 0;
    header.z = 0;
    header.rcode = 0; /* No error */
    header.qdcount = 1;

    /* OBhack: Tack on the header at the beginning without molesting the
       rest of the string */
    length_save = most->unit_count;
    make_hdr(&header,most);
    most->unit_count = length_save;

    /* Add the ar records to the end */
    if(js_append(ar,most) == JS_ERROR) {
	goto giveerror;
	}

    /* Compress "most" and place the compressed data in "ar" */
    if(compress_data(most,ar) == JS_ERROR) {
        js_destroy(ar); 
        udperror(sock,most,client,SERVER_FAIL); 
        js_destroy(most);
        return JS_ERROR;
        }

    /* Check to make sure the data fits in under 512 bytes */
    if(ar->unit_count > 512) {
        /* We handle truncation by truncating everything except the
	   12-byte header */
        header.tc = 1;
        make_hdr(&header,ar);
	}

    /* Success! Put out the good data */
    sendto(sock,ar->string,ar->unit_count,0,client,len_inet);

    js_destroy(most);
    js_destroy(ar);

    return JS_SUCCESS;

    /* We use gotos to make up for C's lack of error trapping */
    giveerror:
        js_destroy(ar);
        udperror(sock,most,client,SERVER_FAIL); 
	js_destroy(most);
	return JS_ERROR;

    }

/* If we have a NXDOMAIN, deliver that record on the udp packet.
   Input: Where a pointer to the rr in question is, the id of the
          query they sent us, the socket the
          UDP bind is on, the sockaddr of the client who sent us the message,
          a js_string containing the query (dname + type)
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int udpnotfound(rr *where,int id,int sock,struct sockaddr *client,
                js_string *query) {
    js_string *most, *compressed; /* Most of the data */
   
    uint16 first_rr_type;
    int length_save;
    int len_inet = sizeof(struct sockaddr);

    q_header header;
    /* Initialize the js_string objects */
    if((most = js_create(1024,1)) == 0)
        return JS_ERROR;
    if((compressed = js_create(1024,1)) == 0) {
        js_destroy(most);
        return JS_ERROR;
        }

    /* Make the header a placeholder for now */
    header.id = id;
    if(make_hdr(&header,most) == JS_ERROR) {
        js_destroy(most); js_destroy(compressed);
        return JS_ERROR;
	}

    /* Sanity check */
    if(where == 0) 
        goto giveerror;
    if(where->query == 0) {
        goto giveerror; 
	}
    if(js_has_sanity(where->query) == JS_ERROR) {
        goto giveerror; 
	}
    if(where->data == 0) {
        goto giveerror; 
	}
    if(js_has_sanity(where->data) == JS_ERROR) {
        goto giveerror;
	}
    if(js_has_sanity(query) == JS_ERROR) {
        goto giveerror;
	}
    first_rr_type = get_rtype(query);

    /* We have to add this header here--authoritative depends on the 
       authorative status of the first record we find */
    header.aa = where->authoritative;

    /* The data must be between 0 and 65535 bytes in length (16-bit
       unsigned value) */
    if(where->data->unit_count < 0 || where->data->unit_count > 65535) {
        goto giveerror;
	}
   
    /* Append the question to the answer */
    if(js_append(query,most) == JS_ERROR) {
        goto giveerror; 
	}

    /* Append the class (in) to the answer */
    if(js_adduint16(most,1) == JS_ERROR) {
        goto giveerror; 
	}

    /* These three values stay at zero */
    header.ancount = 0;
    header.nscount = 1;
    header.arcount = 0;

    /* Append the name for this answer to the answer */
    if(js_append(where->query,most) == JS_ERROR) {
        goto giveerror; 
        }
    /* Append the class (in) to the answer */
    if(js_adduint16(most,1) == JS_ERROR) {
        goto giveerror;
        }
    /* Append the ttl to the answer */
    if(js_adduint32(most,where->ttl) == JS_ERROR) {
        goto giveerror;
        }
    /* Add the rdlength to the answer */
    if(js_adduint16(most,where->data->unit_count) == JS_ERROR) {
        goto giveerror;
        }
    /* Add the record itself to the answer */
    if(js_append(where->data,most) == JS_ERROR) {
        goto giveerror; 
        }

    /* Customize the header */
    /* header.id already set */
    header.qr = 1;
    header.opcode = 0;
    header.tc = 0; /* To do: truncation handling */
    header.rd = 0;
    header.ra = 0;
    header.z = 0;
    /* TO DO: Code that verifies that this host does not exist in 
       any class.  If so, then we give them a rcode of NXDOMAIN_RCODE.
       Otherwise, we give them a rcode of 0 */
    header.rcode = 0; 
    header.qdcount = 1;

    /* OBhack: Tack on the header at the beginning without molesting the
       rest of the string */
    length_save = most->unit_count;
    make_hdr(&header,most);
    most->unit_count = length_save;

    /* Compress "most" and place the compressed data in "compressed" */
    if(compress_data(most,compressed) == JS_ERROR) {
        js_destroy(compressed); 
        udperror(sock,most,client,SERVER_FAIL); 
        js_destroy(most);
        return JS_ERROR;
        }

    /* Check to make sure the data fits in under 512 bytes */
    if(compressed->unit_count > 512) {
        /* We handle truncation by truncating everything except the
	   12-byte header */
        header.tc = 1;
        make_hdr(&header,compressed);
	}

    /* Success! Put out the good data */
    sendto(sock,compressed->string,compressed->unit_count,0,client,len_inet);

    js_destroy(most);
    js_destroy(compressed);

    return JS_SUCCESS;
 

    /* We use gotos to make up for C's lack of error trapping */
    giveerror:
        js_destroy(compressed);
        udperror(sock,most,client,SERVER_FAIL); 
	js_destroy(most);
	return JS_ERROR;

    }
 
/* Given a domain-label starting with a star record ('_') change this label 
   in-place so that the first domain label after the star record is lopped 
   off of it.  Eg. '_\003sub\007example\003com\000" becomes
   "_\007example\003com\000"
   input: A pointer to the js_string object in question
   output: JS_ERROR on error, JS_SUCCESS on success, 0 if the label is
           zero-length already
*/

int bobbit_starlabel(js_string *js) {
    int counter = 1;
    unsigned char length;

    if(js->unit_size != 1)
        return JS_ERROR;
    if(js->unit_count >= js->max_count)
        return JS_ERROR;
    if(js->unit_count < 2)
        return JS_ERROR;
    if(*(js->string) != '_')
        return JS_ERROR;
    length = *(js->string + 1);

    if(length + 2 > js->unit_count || length > 63)
        return JS_ERROR;
    else if(length == 0)
        return 0;

    length++;

    while(counter < (js->unit_count - length) + 1) {
        *(js->string + counter) = *(js->string + counter + length);
	counter++;
	}

    js->unit_count -= length;

    return JS_SUCCESS;

    }

/* Given a domain-label without a star record ('_'), change the first 
   domain label in to a star record ('_') Eg. "\003www\007example\003com\000" 
   becomes "_\007example\003com\000"
   input: A pointer to the js_string object in question
   output: JS_ERROR on error, JS_SUCCESS on success, 0 if the label is
           a star record already
*/

int make_starlabel(js_string *js) {
    int counter = 1;
    unsigned char length;

    if(js->unit_size != 1)
        return JS_ERROR;
    if(js->unit_count >= js->max_count)
        return JS_ERROR;
    if(js->unit_count < 2)
        return JS_ERROR;
    if(*(js->string) == '_')
        return 0;
    length = *(js->string);
    *(js->string) = '_';

    if(length > js->unit_count || length > 63)
        return JS_ERROR;
    if(length == 0) /* We don't convert a "." domain-label */
        return 0; 

    while(counter < js->unit_count - length) {
        *(js->string + counter) = *(js->string + counter + length);
	counter++;
	}

    js->unit_count -= length;

    return JS_SUCCESS;

    }

/* Convert a domain-name query in to its lower-case equivalent 
   Input: Pointer to the js string object with the query
   Output: JS_ERROR on error, JS_SUCCESS on sucess, 0 on 
           success if no change was made to the string */

int fold_case(js_string *js) {
    int counter = 0;
    int ret = 0;

    if(js->max_count <= js->unit_count)
        return JS_ERROR;
    if(js->unit_size != 1)
        return JS_ERROR;
    if(js->unit_count < 2)
        return JS_ERROR;
    while(counter + 2 < js->unit_count) {
        /* Since A-Z never happen in a domain length label, we can speed
	   things up a bit */
        if(*(js->string + counter) >= 'A' && *(js->string + counter) <= 'Z') {
	    *(js->string + counter) += 32;
	    ret = 1;
	    }
        counter++;
	}

    return ret;

    }

/* Check to see if the IP in question is a ddip (e.g. 
   "<03>127<01>0<01>0<03>1<00>"), and, if so, convert it in to 
   a bare A record 
   input: Pointer to js_string object with the query
   output: JS_ERROR on fatal error, 0 on non-ddip query, 
           JS_SUCCESS if it was a ddip
*/

int ddip_check(int id, int sock, struct sockaddr *from, js_string *query) {
    unsigned char ip[4];
    unsigned char length, val;
    int counter, critter, lenl, value;
    js_string *reply;
    q_header header;

    /* Sanity checks */
    if(query->unit_size != 1)
        return JS_ERROR;
    if(query->unit_count >= query->max_count)
        return JS_ERROR;

    /* We presently only do ddip translation for A and ANY requests 
       (DJB only supports this in Dnscache) */
    if(get_rtype(query) != RR_A && get_rtype(query) != RR_ANY) 
        return 0;
   
    if(query->unit_count < 9) /* The minimum possible length for a 
                                 ddip domain label */
        return 0;

    lenl = 0; 
    for(counter=0;counter<4;counter++) {
	length = *(query->string + lenl);
	if(length < 1 || length > 3)
	    return 0;
        critter = lenl + 1;
	lenl += length + 1;
        if(lenl > query->unit_count)
	    return JS_ERROR;
        for(value = 0;critter < lenl;critter++) {
	    val = *(query->string + critter);
	    if(val > '9' || val < '0')
	        return 0;
            value *= 10;
	    value += val - '0';
	    }
        if(value < 0 || value > 255)
	    return 0;
        ip[counter] = value;
	}

    if(*(query->string + lenl) != 0)
        return 0;

    /* OK, it is definitely a ddip label.  Convert the ip in to a DNS reply */

    if((reply = js_create(512,1)) == 0)
        return JS_ERROR;

    /* Build up the header for this reply */
    if(id > 0 && id < 65535)
        header.id = id;
    else 
        goto cleanup;

    header.qr = 1; /* Reply */
    header.opcode = 0; /* Normal DNS */
    header.aa = 0; /* DDIP to A translations are never authoritative */
    header.tc = 0; /* A labels are too short to be truncated */
    header.rd = 0; /* Recursion not desired */
    header.ra = 0; /* Recursion not available */
    header.z = 0; /* This must be 0 unless we are EDNS aware (we aren't) */
    header.rcode = 0; /* Success! */
    header.qdcount = 1;
    header.ancount = 1;
    header.nscount = 0;
    header.arcount = 0;

    /* Make a header of the reply */
    if(make_hdr(&header,reply) == JS_ERROR) 
        goto cleanup;

    /* Add the question they asked to the reply */
    if(js_append(query,reply) == JS_ERROR) 
        goto cleanup;

    /* Add the class (in) to the answer */
    if(js_adduint16(reply,1) == JS_ERROR) 
        goto cleanup;

    /* Make sure the answer is an A record type */
    if(change_rtype(query,RR_A) == JS_ERROR)
        goto cleanup;

    /* We will now add out manufactured A reply */
    if(js_append(query,reply) == JS_ERROR) 
        goto cleanup;
    /* Append the class (in) to the answer */
    if(js_adduint16(reply,1) == JS_ERROR) 
        goto cleanup;
    /* Append a bogus TTL to the answer */
    if(js_adduint32(reply,8675309) == JS_ERROR) 
        goto cleanup;
    /* Add the rdlength to the answer */
    if(js_adduint16(reply,4) == JS_ERROR) 
        goto cleanup;
    /* Add the actual 4-byte reply to the answer */
    for(counter = 0; counter < 4; counter++) {
        if(js_addbyte(reply,ip[counter]) == JS_ERROR) 
	    goto cleanup;
        }

    /* Send the reply out */
    sendto(sock,reply->string,reply->unit_count,0,from,
           sizeof(struct sockaddr));
   
    /* And, we are done */
    js_destroy(reply);
    return JS_SUCCESS;

    /* We use gotos to work around C's lack of error trapping */
    cleanup:
        js_destroy(reply);
	return JS_ERROR;

    }

/* Process the DNS query that comes in from the 'net
   Input: uncompressed form of incoming UDP query, IP address of where
          this query came from, socket number of this socket
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int proc_query(js_string *raw, struct sockaddr *from, int sock) {

    q_header header; /* Header of the question */
    js_string *lookfor; /* What to look for in the big hash */
    js_string *origq; /* Original query asked by the user */
    js_string *lc; /* Lower-case version of query asked by the user */
    rr *nxstore = 0; /* A pointer to the SOA we return when we hit a 
                        NXDOMAIN */
    int length, case_folded, result_code = 0, qtype;
    mhash_e spot_data;
    int have_authority = 0; /* Do we have authority for this record?
                               (must be 1 to return a NXDOMAIN) */
    rr *point;

    /* Sanity checks */
    if(js_has_sanity(raw) == JS_ERROR)
        return JS_SUCCESS;
    if(raw->unit_size != 1)
        return JS_SUCCESS;

    /* Get the header */
    if(read_hdr(raw,&header) == JS_ERROR) { /* Something went wrong,
                                               return error "Format error" */
        udperror(sock,raw,from,FORMAT_ERROR);
        return JS_SUCCESS;
	}

    /* We only support a qdcount of 1 */
    if(header.qdcount != 1) {
        if(no_fingerprint != 1)
            udperror(sock,raw,from,NOT_IMPLEMENTED);
        return JS_SUCCESS;
	}

    /* Get the question from the stream */
    if(raw->unit_count < 14) {
        if(no_fingerprint != 1)
            udperror(sock,raw,from,FORMAT_ERROR);
        return JS_SUCCESS;
	}

    /* Determine the length of the domain label in the question */
    length = dlabel_length(raw,12);
    if(length < 0 || length > 255) {
        if(no_fingerprint != 1)
            udperror(sock,raw,from,FORMAT_ERROR);
        return JS_SUCCESS;
	}

    if(raw->unit_count < 16 + length) { /* 16 because 12 for the header,
                                           and 4 for the type and class */
        if(no_fingerprint != 1)
            udperror(sock,raw,from,FORMAT_ERROR);
        return JS_SUCCESS;
	}

    /* Return "not implemented" if the class is not 1 (Internet class) */
    if(*(raw->string + length + 14) != 0 && 
       *(raw->string + length + 15) != 1) {
        if(no_fingerprint != 1)
            udperror(sock,raw,from,NOT_IMPLEMENTED);
        return JS_ERROR;
	}

    /* Create the lookfor string, returning error if appropriate */
    if((lookfor = js_create(256,1)) == 0) {
        if(no_fingerprint != 1)
            udperror(sock,raw,from,SERVER_FAIL); 
        return JS_ERROR;
	}
    if((origq = js_create(256,1)) == 0) {
        js_destroy(lookfor);
        udperror(sock,raw,from,SERVER_FAIL); 
        return JS_ERROR;
	}
    if((lc = js_create(256,1)) == 0) {
        js_destroy(lookfor); js_destroy(origq);
        udperror(sock,raw,from,SERVER_FAIL); 
        return JS_ERROR;
	}
    if(js_set_encode(lc,JS_US_ASCII) == JS_ERROR) { /* ASCII because we
                                                     only fold the case of 
                                                     A-Z */
        goto serv_fail;
	}
  
    /* Get the query we will look for from their raw query */ 
    if(js_substr(raw,lookfor,12,length + 2) == JS_ERROR) {
        goto serv_fail;
	}

    /* And copy it over to the "original query" */
    if(js_copy(lookfor,origq) == JS_ERROR) {
   	goto serv_fail;
	}

    /* Get the type of query the client desires */
    qtype = get_rtype(origq);
    if(qtype == JS_ERROR) {
        goto serv_fail;
	}

    if(qtype >= 251 && qtype <= 254) { /* AXFR, "All RR types", and 2 more */
        goto not_impl;
	}
  
    /* Handle the case of RR_ANY */
    if(qtype == RR_ANY) {
        udpany(header.id,sock,from,origq);
        js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
	return JS_SUCCESS;
	}
 
    /* OK, start the complicated domain look up routine */
    spot_data = mhash_get(bighash,lookfor);

    /* If found, then give back the appropriate response */
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
        udpsuccess(spot_data.value,header.id,sock,from,origq);
        js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
        return JS_SUCCESS;
	}

    /* If not found, maybe the lowercase version of the same hostname
       resolves. */

    /* Convert to lowercase */
    if(js_copy(lookfor,lc) == JS_ERROR) {
   	goto serv_fail;
	}
    case_folded = fold_case(lc);
    if(case_folded == JS_ERROR) {
   	goto serv_fail;
	}

    /* Look for lower-case version (only if applicable) */
    if(case_folded==1) { /* If this domain label had uppercase letters in it */
        spot_data = mhash_get(bighash,lc);
        if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
            udpsuccess(spot_data.value,header.id,sock,from,origq);
            js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
            return JS_SUCCESS;
	    }
        }

    /* OK, if not found, maybe there is a CNAME with the same domain label */
    if(change_rtype(lookfor,RR_CNAME) == JS_ERROR) {
   	goto serv_fail;
	}
    spot_data = mhash_get(bighash,lookfor); 
   
    /* If the CNAME was found, return the CNAME infomation */
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
        udpsuccess(spot_data.value,header.id,sock,from,origq);
        js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
        return JS_SUCCESS;
	}

    /* Maybe it is the lower-case version of the CNAME */
    if(case_folded==1) { /* If this domain label had uppercase letters in it */
        if(change_rtype(lc,RR_CNAME) == JS_ERROR) {
   	    goto serv_fail;
	    }
        spot_data = mhash_get(bighash,lc); 
        /* If the CNAME was found, return the lower-case CNAME infomation */
        if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
            udpsuccess(spot_data.value,header.id,sock,from,origq);
            js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
            return JS_SUCCESS;
	    }
        }

    /* OK, if not found, maybe there is a *nonauthoritative* NS record with
       the same domain label */
    if(change_rtype(lookfor,RR_NS) == JS_ERROR) {
        goto serv_fail;
	}

    spot_data = mhash_get(bighash,lookfor); 
    point = spot_data.value;
    /* If the non-authoritative NS was found, return the NS infomation */
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR &&
       point->authoritative == 0) {
        udpsuccess(spot_data.value,header.id,sock,from,origq);
        js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
        return JS_SUCCESS;
	}

    /* Maybe a lower-case version of the same ns label */
    if(case_folded==1) { /* If this domain label had uppercase letters in it */
        if(change_rtype(lc,RR_NS) == JS_ERROR) {
   	    goto serv_fail;
	    }

        spot_data = mhash_get(bighash,lc); 
        point = spot_data.value;
        /* If the non-authoritative NS was found, return the NS infomation */
        if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR &&
           point->authoritative == 0) {
            udpsuccess(spot_data.value,header.id,sock,from,origq);
            js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
            return JS_SUCCESS;
	    }
        }

    /* Look for a non-authoritative NS record above.  E.G., if they
       ask for somthing.below.sub.example.com. and we have knowledge
       that sub.example.com is delegated to another nameserver (making
       the NS record in question non-authoritative), return that data
     */
    while(bobbit_label(lookfor) > 0) {
        spot_data = mhash_get(bighash,lookfor);
	point = spot_data.value;
	/* We stop going up the tree if we have found an authoritative NS
	   record */
	if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR &&
	   point->authoritative != 0) {
	    have_authority = 1;
	    /* Look for a SOA record of the same type to prepare for
	       a NXDOMAIN reply */
	    if(change_rtype(lookfor,RR_SOA) == JS_ERROR) {
   		goto serv_fail;
	        }
            spot_data = mhash_get(bighash,lookfor);
            if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
	        nxstore = spot_data.value;
	        } 
	    break;
	    }
        if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
	    udpsuccess(spot_data.value,header.id,sock,from,origq);
            js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
            return JS_SUCCESS;
	    }
	}


    /* Same thing, with case-insensitivity */
    if(case_folded==1) { /* If this domain label had uppercase letters in it */
        while(bobbit_label(lc) > 0) {
            spot_data = mhash_get(bighash,lc);
	    point = spot_data.value;
	    /* We stop going up the tree if we have found an authoritative NS
	       record */
	    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR &&
	       point->authoritative != 0) {
	        have_authority = 1; /* We are allowed to give a NXDOMAIN */
	        /* Look for a SOA record of the same type to prepare for
	           a NXDOMAIN reply */
	        if(change_rtype(lookfor,RR_SOA) == JS_ERROR) {
   		    goto serv_fail;
	            }
                spot_data = mhash_get(bighash,lookfor);
                if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
	            nxstore = spot_data.value;
	            } 
	        break;
		}
            if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
	        udpsuccess(spot_data.value,header.id,sock,from,origq);
                js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
                return JS_SUCCESS;
		}
	    }
        }

    /* If we do not have authority for this record... */
    if(have_authority == 0) {
        /* See if it is a dotted-decimal IP */
        if(no_fingerprint != 1)
	    result_code = ddip_check(header.id,sock,from,origq);
        if(result_code == JS_SUCCESS) {
            js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
	    return JS_SUCCESS;
            }
        if(result_code == JS_ERROR) {
   	    goto serv_fail;
	    }

#ifdef VERSION
	/* A TXT query to "erre-con-erre-cigarro.maradns.org" will
           return the version of MaraDNS being run.  This only
           works if we are not authoritative for "maradns.org", since
           the real "erre-con-erre-cigarro.maradns.org" says
           "MaraDNS version number not available" in the TXT record. */
        if(origq->unit_count == 37 && *(origq->string) == 21
           && no_fingerprint != 1) {
	    result_code = easter_egg(header.id,sock,from,origq,
	       "Terre-con-erre-cigarro.maradns.org.",RR_TXT,"MaraDNS version ",
	       VERSION);
            if(result_code == JS_SUCCESS) {
                js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
	        return JS_SUCCESS;
                }
            if(result_code == JS_ERROR) {
   	        goto serv_fail;
	        }
	    }
#endif /* VERSION */

        /* XXX: Add recurse ACL here */
        /* Launch a separate thread to recursivly determine the 
	   host name in question */
        launch_thread(header.id,sock,*from,origq); 
        js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
	return JS_SUCCESS;
        }

    /* Maybe it is a star record they are looking for */

    /* We need to restore "lookfor" and "lc" because we shredded both
       strings looking for a NS sub-delegation */
    if(js_copy(origq,lookfor) == JS_ERROR) {
        goto serv_fail;
	}
    /* Convert lookfor in to a star label */
    if(make_starlabel(lookfor) == JS_ERROR) {
   	goto serv_fail;
	}

    /* Look for the star record in the big hash */
    spot_data = mhash_get(bighash,lookfor);
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
        udpstar(spot_data.value,header.id,sock,from,origq,origq);
        js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
        return JS_SUCCESS;
	}

    /* Then look for the lowercase star record */

    /* Make a 'lc' (lower case) version of the query if appropriate */
    if(case_folded==1) { /* Naturally, we only deal with lc if we need to */
        if(js_copy(lookfor,lc) == JS_ERROR) {
   	    goto serv_fail;
            }
        if(fold_case(lc) == JS_ERROR) {
   	    goto serv_fail;
	    }
        spot_data = mhash_get(bighash,lc);
	if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
	    /* We have to fold the case of the query answer to lowercase */
	    if(js_copy(origq,lc) == JS_ERROR) {
   	        goto serv_fail;
		}
            if(fold_case(lc) == JS_ERROR) {
   		goto serv_fail;
		}
	    udpstar(spot_data.value,header.id,sock,from,origq,lc);
            js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
            return JS_SUCCESS;
	    }
        }

    /* OK, maybe there is a star record "above".  In other words,
       handle the case when they ask for foo.bar.example.com and we have 
       a record for *.example.com */
    while(bobbit_starlabel(lookfor) > 0) {
        spot_data = mhash_get(bighash,lookfor);
	point = spot_data.value;
	if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
	    udpstar(spot_data.value,header.id,sock,from,origq,origq);
            js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
            return JS_SUCCESS;
	    }
        }

    /* lower-case version of looking for a label "above" */
    if(case_folded==1) {
        while(bobbit_starlabel(lc) > 0) {
	    spot_data = mhash_get(bighash,lc);
	    point = spot_data.value;
	    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
	        /* We have to fold the case of the query answer to lowercase */
	        if(js_copy(origq,lc) == JS_ERROR) {
   		    goto serv_fail;
		    }
                if(fold_case(lc) == JS_ERROR) {
   		    goto serv_fail;
		    }
	        udpstar(spot_data.value,header.id,sock,from,origq,lc);
                js_destroy(lookfor); js_destroy(origq); js_destroy(lc);
                return JS_SUCCESS;
		}
            }
        }

    /* Currently, MaraDNS will not support star records for CNAME nor
       for NS subdelegation.  Code in ParseCsv1.c warns the user of this
       fact. */

    udpnotfound(nxstore,header.id,sock,from,origq);
    js_destroy(lookfor); js_destroy(origq); js_destroy(lc);

    return JS_SUCCESS;

    /* Work around C's lack of error handling and garbage collection with
       gotos */
    serv_fail:
        js_destroy(lookfor); 
	js_destroy(origq); 
	js_destroy(lc);
        if(no_fingerprint != 1)
            udperror(sock,raw,from,SERVER_FAIL); 
	return JS_ERROR;

    not_impl:
        js_destroy(lookfor); 
	js_destroy(origq); 
	js_destroy(lc);
        if(no_fingerprint != 1)
            udperror(sock,raw,from,NOT_IMPLEMENTED);
	return JS_ERROR;

    }

/* Bind to UDP port 53.
   Input:  pointer to socket to bind on, js_string with the dotted-decimal 
           ip address to bind to
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int udpbind(int *sock, js_string *ip_ddip) {
    int len_inet; /* Length */
    struct sockaddr_in dns_udp;
    char ip[64];

    /* Sanity checks */
    if(sock == 0)
        return JS_ERROR;
    if(js_has_sanity(ip_ddip) == JS_ERROR)
        return JS_ERROR;
    if(ip_ddip->unit_size != 1)
        return JS_ERROR;

    /* Create string to hold IP */
    if(js_js2str(ip_ddip,ip,60) == JS_ERROR)
        return JS_ERROR;

    /* Create a raw UDP socket */
    if((*sock = socket(AF_INET,SOCK_DGRAM,0)) == -1) {
	return JS_ERROR;
	}
   
    /* Create a socket address to use with bind() */
    memset(&dns_udp,0,sizeof(dns_udp));
    dns_udp.sin_family = AF_INET;
    dns_udp.sin_port = htons(53);
    if((dns_udp.sin_addr.s_addr = inet_addr(ip)) == INADDR_NONE)
        return JS_ERROR;

    len_inet = sizeof(dns_udp);

    /* Bind to the socket.  Note that we usually have to be root to do this */
    if(bind(*sock,(struct sockaddr *)&dns_udp,len_inet) == -1)
        return JS_ERROR;

    /* We are now bound to UDP port 53.  Leave */
    return JS_SUCCESS;
    }

/* Get information from a previously binded UDP socket
   Input:  UDP bound socket, pointer to sockaddr structure that will contain
           the IP of the system connecting to us, pointer to js_string
           object that will have the data in question, maximum allowed
           length of data we receive
   Output: JS_ERROR on error, JS_SUCCESS on success 
*/

int getudp(int sock,struct sockaddr *client, js_string *data, int max_len) {
    int len_inet, counter;

    /* Sanity checks */
    if(client == 0 || data == 0)
        return JS_ERROR;
    if(js_has_sanity(data) == JS_ERROR)
        return JS_ERROR;
    if(data->unit_size != 1)
        return JS_ERROR;
    if(max_len < 0 || max_len >= data->max_count)
        return JS_ERROR;

    len_inet = sizeof(struct sockaddr);

    counter = recvfrom(sock,data->string,max_len,0,client,&len_inet);

    if(counter < 0)
        return JS_ERROR;

    data->unit_count = counter;

    return JS_SUCCESS;

    }

/* The core of the DNS server */

int main(int argc, char **argv) {
   
    js_string *mararc_loc, *errors, *chrootn, *uidstr, *maxpstr,
              *kvar_query, *bind_address, *incoming, *uncomp, *verbstr;
    unsigned char chroot_zt[255];
    int errorn, value, uid, sock, maxprocs;
    struct sockaddr client;
    struct rlimit rlim;

    /* Initialize the strings (allocate memory for them, etc.) */
    if((mararc_loc = js_create(256,1)) == 0)
        harderror(L_MLC); /* "Could not create mararc_loc string" */
    if(js_set_encode(mararc_loc,MARA_LOCALE) == JS_ERROR)
        harderror(L_MLL); /* "Could not set locale for mararc_loc string" */
    if((errors = js_create(256,1)) == 0)
        harderror(L_EC); /* "Could not create errors string" */
    if(js_set_encode(errors,MARA_LOCALE) == JS_ERROR)
        harderror(L_EL); /* "Could not set locale for errors string" */
    if((uidstr = js_create(256,1)) == 0)
        harderror(L_UC); /* "Could not create uidstr string" */
    if(js_set_encode(uidstr,MARA_LOCALE) == JS_ERROR)
        harderror(L_UL); /* "Could not set locale for uidstr string" */
    if((verbstr = js_create(256,1)) == 0)
        harderror(L_VC); /* "Could not create verbstr string" */
    if(js_set_encode(verbstr,MARA_LOCALE) == JS_ERROR)
        harderror(L_VL); /* "Could not set locale for verbstr string" */
    if((maxpstr = js_create(256,1)) == 0)
        harderror(L_MC); /* "Could not create maxpstr string" */
    if(js_set_encode(maxpstr,MARA_LOCALE) == JS_ERROR)
        harderror(L_ML); /* "Could not set locale for maxpstr string" */
    if((chrootn = js_create(256,1)) == 0)
        harderror(L_CC); /* "Could not create chrootn string" */
    if(js_set_encode(chrootn,MARA_LOCALE) == JS_ERROR)
        harderror(L_CL); /* "Could not set locale for chrootn string" */
    if((kvar_query = js_create(256,1)) == 0)
        harderror(L_KQC); /* "Could not create kvar_query string" */
    if(js_set_encode(kvar_query,MARA_LOCALE) == JS_ERROR)
        harderror(L_KQL); /* "Could not set locale for kvar_query string" */
    if((bind_address = js_create(64,1)) == 0)
        harderror(L_BAC); /* "Could not create bins_address string" */
    if(js_set_encode(bind_address,MARA_LOCALE) == JS_ERROR)
        harderror(L_BAL); /* "Could not set locale for bind_address string" */
    if((incoming = js_create(768,1)) == 0)
        harderror(L_IC); /* "Could not create incoming string" */
    if(js_set_encode(incoming,MARA_LOCALE) == JS_ERROR)
        harderror(L_IL); /* "Could not set locale for incoming string" */
    if((uncomp = js_create(768,1)) == 0)
        harderror(L_UCC); /* "Could not create uncomp string" */
    if(js_set_encode(uncomp,MARA_LOCALE) == JS_ERROR)
        harderror(L_UCL); /* "Could not set locale for uncomp string" */

    /* First, find the mararc file */
    if(argc == 1) { /* No arguments */
        if(find_mararc(mararc_loc) == JS_ERROR)
            harderror(L_LOC_MARARC); /* "Error locating mararc file" */
        }
    else if(argc==3) { /* maradns -f /wherever/mararc */
        if(js_qstr2js(mararc_loc,argv[2]) == JS_ERROR)
            harderror(L_MARARC_ARG); /* "Could not get mararc from command line" */
        }
    else
        harderror(L_USAGE); /* "Usage: mararc [-f mararc_location]" */

    /* Then parse that file */
    if(read_mararc(mararc_loc,errors,&errorn) == JS_ERROR)
        harderror(L_MARARC_PARSE); /* "Error parsing contents of mararc file" */
    if(errorn != 0) {
        if(errorn != -1)
          /* "Error parsing contents of mararc file on line " */
          printf("%s%d%s",L_MARARC_LINE,errorn,L_N); /* errorn, "\n" */
	printf("%s",L_ERROR_CODE); /* "Error code: " */
	js_show_stdout(errors);
	printf("%s",L_N); /* "\n" */
	exit(2);
	}

    /* Get in to a state of least privledge ASAP */

    /* Limit the maximum number of processes */
    if(js_qstr2js(kvar_query,"maxprocs") == JS_ERROR)
        harderror(L_KVAR_Q); /* "Could not create kvar_query" */
    if(read_kvar(kvar_query,maxpstr) == JS_ERROR)
        harderror(L_MAXPROC); /* "Problem getting maxprocs value.\nmaxprocs must be set before starting the MaraDNS server" */
    if((maxprocs = js_atoi(maxpstr,0)) == 0)
        harderror(L_MAXPROC_NUM); /* "Problem converting maxprocs to a number\nThis must be a non-zero number" */
    rlim.rlim_cur = rlim.rlim_max = maxprocs;
    if(setrlimit(RLIMIT_NPROC,&rlim) != 0)
        sys_harderror(L_MAXPROC_SET); /* "Unable to set maximum number of processes" */

    /* Determine the level of error reporting */
    if(js_qstr2js(kvar_query,"verbose_level") == JS_ERROR)
        harderror(L_KVAR_Q); /* "Could not create kvar_query" */
    if(read_kvar(kvar_query,verbstr) == JS_ERROR)
        log_level = 0;
    else
        log_level = js_atoi(verbstr,0);
    /* Determine if we should make a "best faith" effort to have
       no MaraDNS-specific features */
    if(js_qstr2js(kvar_query,"no_fingerprint") == JS_ERROR)
        harderror(L_KVAR_Q);
    if(read_kvar(kvar_query,verbstr) == JS_SUCCESS) {
        no_fingerprint = js_atoi(verbstr,0);
        }
        
    /* Determine if we are root */
    if(geteuid() == 0) {
        /* Change the root directory */
        if(js_qstr2js(kvar_query,"chroot_dir") == JS_ERROR)
	    harderror(L_KVAR_Q); /* "Could not create kvar_query" */
        if(read_kvar(kvar_query,chrootn) == JS_ERROR)
	    harderror(L_CHROOT_KVAR); /* "Problem getting chroot kvar.\nYou must have chroot_dir set if you start this as root" */
        if(js_js2str(chrootn,chroot_zt,200) == JS_ERROR)
	    harderror(L_CHROOT_NT); /* "Problem making chroot nt string.\nMake sure the chroot directory is 200 chars or less" */
        if(chdir(chroot_zt) != 0)
	    sys_harderror(L_CHROOT_CHANGE); /* "Problem changing to chroot dir.\nMake sure chroot_dir points to a valid directory" */
        if(chroot(chroot_zt) != 0)
	    sys_harderror(L_CHROOT_DO);  /* "Problem changing the root directory." */
  
        mlog(L_CHROOT_SUCCESS); /* "Root directory changed" */
     
	/* Bind to port 53 
           To Do: use capset to give us privledged bind abilities without 
                  needing to be root.
        */
        if(js_qstr2js(kvar_query,"bind_address") == JS_ERROR)
	    harderror(L_KVAR_Q); /* "Could not create kvar_query" */
        if(read_kvar(kvar_query,bind_address) == JS_ERROR)
	    harderror(L_NOBIND); /* "Problem getting chroot kvar.\nYou must have bind_address set to the IP maradns will listen on" */
        if(udpbind(&sock,bind_address) == JS_ERROR)
            sys_harderror(L_BINDFAIL); /* "Problem binding to port 53.\nMost likely, another process is already listening on port 53" */

        mlog(L_BIND_SUCCESS);  /* "Socket opened on UDP port 53" */

        /* Drop the elevated privledges */
	if(js_qstr2js(kvar_query,"maradns_uid") == JS_ERROR)
	    harderror(L_KVAR_Q); /* "Could not create kvar_query" */
        if(read_kvar(kvar_query,uidstr) == JS_ERROR)
	    harderror(L_NOUID); /* "Problem getting maradns_uid kvar.\nYou must have maradns_uid set if you start this as root" */
        if((uid = js_atoi(uidstr,0)) < 10)
	    harderror(L_BADUID); /* "maradns_uid is less than 10 or not a number.\nThis uid must have a value of 10 or more" */
        if(setuid(uid) != 0)
	    sys_harderror(L_NODROP); /* "Could not drop root uid" */
        if(setuid(0) == 0)
	    sys_harderror(L_STILL_ROOT);  /* "We seem to still be root" */

        mlog(L_DROP_SUCCESS); /* "Root privledges dropped" */

        }
    else {
        
	/* Bind to port 53 as a non-root user */
        if(js_qstr2js(kvar_query,"bind_address") == JS_ERROR)
	    harderror(L_KVAR_Q); /* "Could not create kvar_query" */
        if(read_kvar(kvar_query,bind_address) == JS_ERROR)
	    harderror(L_NOBIND); /* "Problem getting chroot kvar.\nYou must have bind_address set to the IP maradns will listen on" */
        if(udpbind(&sock,bind_address) == JS_ERROR)
            sys_harderror(L_BEROOT); /* "Problem binding to port 53.\nYou should run this as root" */
        mlog(L_BIND_SUCCESS);  /* "Socket opened on UDP port 53" */
        }

    /* Create the big hash */
    bighash = 0;
    bighash = mhash_create(8);
    if(bighash == 0)
        harderror(L_NOBIGHASH); /* "Could not create big hash" */
 
    value = populate_main(bighash,errors);
    if(value == JS_ERROR)
        harderror(L_NOPOPULATE); /* "Error running populate_main program" */
    else if(value == -2) {
        js_show_stdout(errors);
	printf("%s",L_N); /* "\n" */
	harderror(L_POPULATE_FATAL); /* "This error in populate hash is fatal" */
	}

    mlog(L_RRS_LOADED);  /* "All RRs have been loaded" */

    /* Load the "seed" data in to the DNS cache */
    if(init_cache() == JS_ERROR)
        harderror(L_INITCACHE_FAIL); /* "Init_cache() failed" */

    /* Enable signal handler for SIGPIPE */
    signal(SIGPIPE,handle_sigpipe);
    /* Some other catchable signals */
    signal(SIGHUP,handle_signal); /* Eventually, reload the zones */
    signal(SIGALRM,handle_signal); 
    signal(SIGUSR1,handle_signal); 
    signal(SIGUSR2,handle_signal); 
   
    /* Listen for data on the UDP socket */
    for(;;) { 
        if(log_level >= 3)
            mlog(L_DATAWAIT); /* "Awaiting data on port 53" */
        if(getudp(sock,(struct sockaddr *)&client,incoming,512) == JS_ERROR)
	    continue;
        if(log_level >= 3)
            mlog(L_GOTDATA);	 /* "Message received, processing" */
	if(decompress_data(incoming,uncomp) == JS_ERROR)
	    continue;
        proc_query(uncomp,(struct sockaddr *)&client,sock); 
        }
    
    }
    
