/* Placed in the public domain 2001 by Sam Trenholme */

/* This is the only code that uses threads */
#include <pthread.h>

/* The MaraDNS includes */
#include "../libs/MaraHash.h"
#include "../MaraDns.h"
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "../dns/functions_dns.h"
#include "../parse/functions_parse.h"
#include "functions_server.h"

pthread_mutex_t big_lock = PTHREAD_MUTEX_INITIALIZER;
#define read_lock() pthread_mutex_lock(&big_lock)
#define read_unlock() pthread_mutex_unlock(&big_lock)
#define write_lock() pthread_mutex_lock(&big_lock)
#define write_unlock() pthread_mutex_unlock(&big_lock)

mhash *dnscache = 0;

typedef struct {
    int id;
    int sock;
    struct sockaddr client;
    js_string *query;
    } dnsreq; /* DNS request */

/* DNS server that is closer to the answer for a given request */
typedef struct closer {
    unsigned int num_elements; /* Number of elements in the linked list */
    time_t ttd; /* Time for this record to die */
    int datatype; /* This is the RR type of the data (RR_A or RR_NS) */
    void *data;
    struct closer *masked; /* If another record with a ttd after the
                              ttd for this record exists, we want to 
			      access that data when this record expires */
    struct closer *next; } closer;

/* The final argument to the functions that add elements to the cache */
#define OVERWRITE 1
#define APPEND 2

#ifdef DEBUG
/* Debug routine that shows an IP in dotted decimal format on the
   standard output.
   Input:  A uint32 ip 
   Output: none
*/
void debug_show_ip(uint32 ip) {
    printf("%d.%d.%d.%d",(ip & 0xff000000) >> 24,
                         (ip & 0x00ff0000) >> 16,
                         (ip & 0x0000ff00) >>  8,
                          ip & 0x000000ff);
    }

/* Debug routine to browse the contents of an assosciate array (dictionary,
   hash table)
   Input: Pointer to hash table
   Output: none
*/

void debug_explore_dict(mhash *dict) {
     /* Lots of cut and paste from mtest.c */
     char strn[256];
     mhash_offset counter,number;
     closer *close;

     for(;;) {
         printf("s to see the entire hash, v for verbose view, q to resume ");
	 printf("execution of maradns\n");
	 printf("Enter command: ");
	 fgets(strn,200,stdin);
	 if(strn[strlen(strn) - 1] == '\n')
	     strn[strlen(strn) - 1] = '\0';
         if(*strn == 's' || *strn == 'v') {
	     printf("Viewing the assosciative array\n");
	     /* Make number 2^hash_bits */
	     number = 1;
	     for(counter = 0; counter < dict->hash_bits; counter++)
                 number *= 2;
             for(counter = 0; counter < number; counter++) {
	         if(dict->hash_table[counter] != 0 &&
		    dict->hash_table[counter]->key != 0 &&
		    js_has_sanity(dict->hash_table[counter]->key)) {
		     printf("Key: ");
		     show_esc_stdout(dict->hash_table[counter]->key);
                     printf("\nValue: ");
		     if(dict->hash_table[counter]->datatype == MARA_DNS_NS) {
		         printf("Name server");
			 if(*strn == 'v') {
			     close = dict->hash_table[counter]->value;
			     printf("\nNum_elements: %d", close->num_elements);
			     printf(" Time to die: %d", close->ttd);
			     printf(" Datatype: %d", close->datatype);
			     if(close->datatype == RR_A) {
			         printf(" A Data: ");
				 debug_show_ip(*(uint32 *)(close->data));
				 }
                             printf("\n");
			     }
			 }
                     else if(dict->hash_table[counter]->datatype == MARA_DNSRR)
		         printf("DNS resource record");
                     else
		         printf("Non-DNS datatype (why is this here?)");
                     printf("\n\n");
		     }
                 } 	
             }
         else if(*strn == 'q') 
	     break;
	 }    
    }               
#endif /* DEBUG */

/* A cryptgraphically secure random number with a value between
   0 and 65535
   Input: None
   Output: A number between 0 and 65535
*/
uint16 srng() {
     /* XXX make this secure */
     return 1234;
     }

/* Determine if a host label with is a substring of a long js
   object is in bailiwick.
   Input: String with host name, offset of beginning of host name,
          bailiwick we are allowed to be in
   Output: JS_ERROR (-1) on fatal error, 0 if host name is out
           of bailiwick, 1 if host name is in bailiwick 
*/

int offset_bailiwick(js_string *js, int offset, js_string *bailiwick) {
    js_string *sub;
    int len, ret, len_save;

    /* Sanity check */
    if(bailiwick->unit_size != 1)
        return JS_ERROR;

    /* Determine how long the domain label is */
    len = dlabel_length(js,offset);
    if(len == JS_ERROR)
        return JS_ERROR;

    /* Make a string which holds just the domain name label */
    if((sub = js_create(len + 1,1)) == 0)
        return JS_ERROR;
    if(js_substr(js,sub,offset,len) == JS_ERROR) {
        js_destroy(sub);
	return JS_ERROR;
	}

    /* Just in case Bailiwick is longer than the dlabel at the top of the
       Bailiwick, truncate the bailiwick */
    len_save = bailiwick->unit_count;
    len = dlabel_length(bailiwick,0);
    if(len == JS_ERROR) {
        js_destroy(sub);
	return JS_ERROR;
	}
    bailiwick->unit_count = len;
    /* Run in_bailiwick with the substring */
    ret = in_bailiwick(sub,bailiwick);
    bailiwick->unit_count = len_save;
#ifdef DEBUG
    printf("B: %d\n",ret);
    show_esc_stdout(sub);
    printf("\n");
    show_esc_stdout(bailiwick);
    printf("\nBailiwick\n");
#endif /* DEBUG */
    js_destroy(sub);
    return ret;
    }

/* Determine if a given host name is in the bailiwick of another 
   domain name.
   Input: Host name we are querying, bailiwick we are allowed to be in
   Output: JS_ERROR (-1) on fatal error, 0 if host name is out
           of bailiwick, 1 if host name is in bailiwick 
*/

int in_bailiwick(js_string *host, js_string *bailiwick) {
    js_string *get;
    int result;
    if((get = js_create(host->unit_count + 3,1)) == 0)
        return JS_ERROR;
    if(js_copy(host,get) == JS_ERROR)
        return JS_ERROR;
    get->encoding = bailiwick->encoding;
    result = 0; /* Out of bailiwick */
    if(js_issame(get,bailiwick) == 1)
        return 1; /* In bailiwick */
    while(result == 0 && get->unit_count > bailiwick->unit_count) {
        if(bobbit_label(get) == JS_ERROR)
	    return JS_ERROR;
	if(js_issame(get,bailiwick) == 1)
	    return 1; /* In bailiwick */
        }
    return result; /* Out of bailiwick */
    }

/* Given two pointers at differnet places in the same js_string object,
   both of which point to a dns-style domain name (in binary format),
   determine if the two objects point to the same domain name.
   Input: Pointer to the JS string we are doing the comparison of, index of
          the first point we start our comparison at, index of the second
	  point we start our comparison at.
   Output: JS_ERROR (-1) on fatal error, 0 if they do not match, 1 if
           they do match */
int cmp_dnames(js_string *js, int p1, int p2) {
    int temp, length;

    if(js_has_sanity(js) == JS_ERROR)
        return JS_ERROR;

    /* Make sure p1 is smaller than p2 */
    if(p1 > p2) {
        temp = p1;
	p1 = p2;
	p2 = temp;
	}
  
    /* If they point to the same place, they are the same (obviously) */
    if(p1 == p2)
        return 1;

    /* If p1 or p2 are out of bounds, then return error 
       Optimization: p2 is greater then p1, so we only have to check to see
       if p1 is less then zero */
    if(p1 < 0 || p1 > js->unit_count || p2 > js->unit_count)
        return JS_ERROR;

    length = dlabel_length(js,p1);

    /* They are not the same if they overlap */
    if(p1 + length > p2)
        return 0;

    /* They are not the same if they have different lengths */
    if(length != dlabel_length(js,p2))
        return 0;

    /* Make sure we are in bounds (optimization: we know p2 is greater than
       p1) */
    if(p2 + length > js->unit_count)
        return JS_ERROR;

    /* OK, now compare the two dlabels byte by byte */
    temp = 0;
    while(temp < length) {
        if(*(js->string + p1 + temp) != *(js->string + p2 + temp))
	    return 0;
        temp++;
        }
    
    /* There is no difference between the labels.  Return 1 */
    return 1;

    }

/* Query a nameserver for the answer to our question.  Add the answer
   (it being a closer nameserver or the question being asked)
   If answer found, add it to the cache.  If pointer to other name servers 
   given, add those pointers to this cache. 
   Input: IP of server we will contact, query to send to the server,
          bailiwick of the IP of the server we will contact
   Output: JS_ERROR on fatal error, -2 if the nameserver could not
           be contacted (timeout), -3 if the namserver refuses to give us the
	   the appropriate information, -4 if it is a lame delegation
	   (non-authoritative data), JS_SUCCESS if we find a name server,
	   2 if we got the final answer
*/

int query_nameserver(int remote_ip, js_string *query, js_string *bailiwick) {
    /* Lots of cut and paste from askmara.c here */
    struct sockaddr_in dns_udp, server;
    int len_inet; /* Length */
    int s; /* Socket */
    int q; /* Used for binding, etc */
    int nspoint, offset, len;
    int type, class, rdlength;
    uint32 ttl;
    int record_added = 0; /* Boolean: has a record been added to the cache */
    js_string *outdata; /* Outgoing data */
    js_string *indata, *uindata; /* Incoming data (uncompressed version) */
    js_string *jsip; /* Place where we put the IP of the answer */
    q_header header; /* header data */
    int counter,place,count;
    /* Select() [which we are using *only* because of its timeout ability]
       requires four variables to be used */
    fd_set rx_set; int maxd; struct timeval tv; int n;
    /* We keep track of what NS servers are servers for a given
       branch of the "domain" tree in the answer */
    uint16 ns_record[32]; /* Who is going to have more than 32 NS records 
                             for a given domain name? */
    /* We keep track of that branch of the DNS "tree" the records are
       authoritative for in the answer */
    uint16 ns_domain[32];

    /* Initialize ns_record and ns_domain */
    ns_record[0] = 0;
    ns_domain[0] = 0;

    /* Allocate memory for some strings */
    if((indata = js_create(512,1)) == 0)
        return JS_ERROR;
    if((uindata = js_create(2048,1)) == 0) {
        js_destroy(indata);
        return JS_ERROR;
	}
    if((outdata = js_create(512,1)) == 0) {
        js_destroy(indata);
	js_destroy(uindata);
        return JS_ERROR;
	}

    /* Create a UDP client socket */
    if((s = socket(AF_INET,SOCK_DGRAM,0)) == -1)
        goto cleanup;

    /* Create a socket address to use with sendto() */
    memset(&dns_udp,0,sizeof(dns_udp));
    dns_udp.sin_family = AF_INET;
    /* XXX read some kvars which determine the range of this,
           also, find out how to change the local port number */
    /*dns_udp.sin_port = htons(20000 + (srng() & 4095)); */
    dns_udp.sin_port = htons(53); 
    if((dns_udp.sin_addr.s_addr = htonl(remote_ip)) == INADDR_NONE)
        return JS_ERROR;

    len_inet = sizeof(dns_udp);

    /* Format a DNS request */
    header.id = srng();
    header.qr = 0;
    header.opcode = 0;
    header.aa = 0;
    header.tc = 0;
    header.rd = 1;
    header.ra = 0;
    header.z = 0;
    header.rcode = 0;
    header.qdcount = 1;
    header.ancount = 0;
    header.nscount = 0;
    header.arcount = 0;

    /* Make a beginning of a DNS query from that header */
    if(make_hdr(&header,outdata) == JS_ERROR)
        goto cleanup;
    if(js_append(query,outdata) == JS_ERROR)
        goto cleanup;
    if(js_adduint16(outdata,1) == JS_ERROR) /* Adding class to query */
        goto cleanup;

    /* OK, the DNS request has been formed.  Send it out */
    if(sendto(s,outdata->string,outdata->unit_count,0,
       (struct sockaddr *)&dns_udp,len_inet) < 0)
        goto cleanup;

    /* Wait for a reply from the DNS server */
    FD_ZERO(&rx_set);
    FD_SET(s,&rx_set);
    maxd = s + 1;
    tv.tv_sec = 2;
    tv.tv_usec = 0;
    n = select(maxd,&rx_set,NULL,NULL,&tv);
    if(n == -1)  /* select error */
        goto cleanup;
    if(n == 0) /* Timeout */
        return -2;
    /* Get the actual reply from the DNS server */
    if((count = recvfrom(s,indata->string,indata->max_count,0,
                         (struct sockaddr *)&dns_udp,&len_inet)) < 0) {
        return -3;
        }
    /* Process the reply */
    indata->unit_count = count;
    decompress_data(indata,uindata);
    if(read_hdr(uindata,&header) == JS_ERROR) 
        return -3;

#ifdef DEBUG
    show_esc_stdout(uindata);
    printf("\nheader.qdcount: %d\n",header.qdcount);
    printf("header.ancount: %d\n",header.ancount);
    printf("header.nscount: %d\n",header.nscount);
    printf("header.arcount: %d\n",header.arcount);
#endif /* DEBUG */
    
    /* Determine if this is an answer or a referral */
    if(header.ancount > 0) {
        /* XXX: If this is an answer, add the answer to our cache */
	return -2; /* REMOVE when code added */
	}
   
    /* OK, it is a referral */
    
    /* If there are no authority records either, then it is an error */
    if(header.nscount == 0) 
        return -2; 

    /* OK, let us find the offset of the first answer */
    count = header.qdcount; 
    offset = 12; /* Point it at the first question/answer */
    while(count > 0) {
        len = dlabel_length(uindata,offset);
        if(len == JS_ERROR)
            goto cleanup;
        offset += len;
        offset += 4; /* Move past the query type and query class */
	count--;
	}

#ifdef DEBUG
    printf("At end of questions, offset is %d byte is %o\n",offset,
           *(uindata->string + offset));
#endif /* DEBUG */

    /* Offset now points at the first authority (NS) record */
    count = header.nscount;
    nspoint = 0;
    /* We only look at the first 29 NS records */
    if(count > 29)
       count = 29;

    /* Start filling up the ns_record array with pointers to the
       various authority records that are ns_records.  Later on,
       we will see which ns records have corresponding records in
       the additional records section. */
       
    nspoint = 0;
    while(nspoint < count) {
        /* Determine the length of the answer returned */
        len = dlabel_length(uindata,offset);
	if(len == JS_ERROR)
	    goto cleanup;
        /* Point to the question they asked 
	   (but only if this is a NS record) */
        /* XXX: Handle SOA records in the NS section */
	if(js_readuint16(uindata,offset + len) == RR_NS && offset > 0)
            ns_domain[nspoint] = offset;
        else {
	    ns_domain[nspoint] = 1; /* Invalid data */
#ifdef DEBUG
            printf("442\n");
#endif /* DEBUG */
            }
        /* Answers that are out of bailiwick are invalid */
	if(offset_bailiwick(uindata,offset,bailiwick) != 1) {
	    ns_domain[nspoint] = 1; /* Invalid data */
#ifdef DEBUG
            printf("449\n");
#endif /* DEBUG */
            }
        ns_domain[nspoint + 1] = 0; /* End of ns records */
        offset += len + 8; /* Go past name, type, class, and ttl */
	len = js_readuint16(uindata,offset); /* rdlength */
	if(len == JS_ERROR)
	    goto cleanup;
        /* We also keep a note of what the NS record is for
	   this domain */
        if(ns_domain[nspoint] > 2) { /* If this is a valid NS record */
            ns_record[nspoint] = offset + 2;
	    }
        else {
	    ns_record[nspoint] = 1; /* invalid data */
#ifdef DEBUG
            printf("465\n");
#endif /* DEBUG */
	    }
	ns_record[nspoint + 1] = 0;
	offset += len + 2; /* Go past rdlength and rddata */
        nspoint++; 
        }
#ifdef DEBUG
    printf("End of NS records, nspoint is %d offset is %d (byte %o)\n",nspoint,
           offset,*(uindata->string + offset));
#endif /* DEBUG */

    /* This string is used to pass arguments to add_closer_jsip */
    if((jsip = js_create(6,1)) == 0)
        goto cleanup;

    /* Start looking at the additional records.  See which ones are IPs
       which match ns records.  When we see such an IP, we add the record
       to the the big cache. */

    count = header.arcount;
    while(count > 0) {
        /* Determine the length of the question we are answering */
	len = dlabel_length(uindata,offset);
	if(len == JS_ERROR)
	    goto cleanup_jsip;
        /* We will only add this record to the cache if:
	   1) The answer is a NS record 
	   2) The answer is in bailiwick (check done above)
	   3) There is a ns record above corresponding to the answer
         */
	type = js_readuint16(uindata,offset + len);
	class = js_readuint16(uindata,offset + len + 2);
	ttl = js_readuint32(uindata,offset + len + 4);
	rdlength = js_readuint16(uindata,offset + len + 8);
	if(type == JS_ERROR || class == JS_ERROR || ttl == JS_ERROR || 
	   rdlength == JS_ERROR)
	    goto cleanup_jsip;
        /* Move the offset pointer to point to the beginning of the
	   rddata */
        if(type != RR_A || class != 1 || rdlength != 4) {
	    offset += len + rdlength + 10; /* Rdlength plus ten byte header,
	                                      plus length of domain name
					      label */
            count--;
	    continue;
	    }
        /* Start comparing the dlabel for this RR with the various NS
	   RRs in the authority section */
        nspoint = 0;
	while(ns_record[nspoint] != 0) {
#ifdef DEBUG
	    printf("Nspoint: %d\n",nspoint);
#endif /* DEBUG */
	    /* Give the jsip string a value */
	    if(js_substr(uindata,jsip,len + offset + 10,4) == JS_ERROR)
	        goto cleanup_jsip;
            /* If the domain name label for this node share the same
	       name as a ns record */
#ifdef DEBUG
	    printf("cmp_dnames: %d (%o) %d (%o)\n",ns_record[nspoint],
	           *(uindata->string + ns_record[nspoint]), offset,
		   *(uindata->string + offset));
#endif /* DEBUG */
	    if(ns_record[nspoint] != 1 && 
	        cmp_dnames(uindata,ns_record[nspoint],offset) == 1) {
	        if(record_added == 0) {
	            record_added = 1;
#ifdef DEBUG
		    printf("DEBUG: %d ",ns_domain[nspoint]);
		    show_esc_stdout(jsip);
		    printf(" OVERWRITE\n");
#endif /* DEBUG */
		    if(add_closer_jsip_offset(uindata,ns_domain[nspoint],
		                              jsip,OVERWRITE) == JS_ERROR)
		        goto cleanup_jsip;
		    }
                else {
#ifdef DEBUG
		    printf("DEBUG: %d ",ns_domain[nspoint]);
		    show_esc_stdout(jsip);
		    printf(" APPEND\n");
#endif /* DEBUG */
		    if(add_closer_jsip_offset(uindata,ns_domain[nspoint],
		                              jsip,APPEND) == JS_ERROR)
		        goto cleanup_jsip;
                    }
                }
            nspoint++;
	    }
        offset += len + rdlength + 10;
	count--;
        }

    js_destroy(jsip);
    js_destroy(indata);
    js_destroy(uindata);
    js_destroy(outdata);
#ifdef DEBUG
    debug_explore_dict(dnscache);
#endif /* DEBUG */
    return JS_SUCCESS;

    cleanup_jsip:
        js_destroy(jsip);
    cleanup:
        js_destroy(indata);
	js_destroy(uindata);
	js_destroy(outdata);
	return JS_ERROR;
    }

/* Recurse thread: This is a special routine that launches its own thread 
   input: Pointer to dnsreq query
   output: JS_SUCCESS (never looked at--this is a detached thread)
*/

int recurse_thread(dnsreq *req) {

    int ret;

    ret = recurse_call(req->id, req->sock, req->client, req->query, 0, 0);
    js_destroy(req->query);
    js_dealloc(req);
    return ret;

    }

/* We made this a separate function so that we can call it recursivly 
   without having to allocate memory on the heap for req objects */

int recurse_call(int id, int sock, struct sockaddr client, 
                 js_string *query, int recurse_depth, int glueless_level) {

    mhash_e spot_data;
    js_string *copy;
    closer *cpoint;
    uint32 nsip;
    int result, qtype;

    /* Make sure we haven't overloaded the recursion depth or maximum glueless 
       level */
    if(recurse_depth > 32 || glueless_level > 4)
        return JS_ERROR;

    /* Look for query in cache */
    read_lock();
    spot_data = mhash_get(dnscache,query); 
 
    /* Return cached data if found */
    
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNSRR) {
        /* XXX: Make sure the RR has not expired */
        udpsuccess(spot_data.value,id,sock,&client,query);
        read_unlock();
	return JS_SUCCESS;
	}

    read_unlock();

    /* OK, we don't have data in the cache.  See which nameserver we
       have to bug */
    if((copy = js_create(257,1)) == 0) 
        goto cleanup_nojs;
    if(js_copy(query,copy) == JS_ERROR)
        goto cleanup_nojs;
   
    /* See if there is a name server in the cache for the exact query we are
       asking for */
    if(change_rtype(copy,252) == JS_ERROR)
        goto cleanup;

    /* OK, look for the "lowest" branch in the tree which will lead us to 
       the data we are seeking */
    do {
        read_lock();
	spot_data = mhash_get(dnscache,copy);
#ifdef DEBUG
	show_esc_stdout(copy); printf("\n"); /*DEBUG*/
#endif /* DEBUG */
	/* Query nameserver if found */
	if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NS) {
	    cpoint = spot_data.value;
	    /* To do: code that makes a local copy of the entire 
	       cpoint chain, so that other threads can destroy the
	       original XXX */
	    /* For now, we just look for the first NS RR with an IP
	       as opposed to a name.  */
	    while(cpoint->datatype != RR_A && cpoint != NULL)
	        cpoint = cpoint->next;
            if(cpoint == NULL)
                return JS_ERROR; /* XXX: Look up the name of an oobaliwick
		                    NS entry */
            nsip = *(uint32 *)(cpoint->data);
	    read_unlock();
            /* Query the nameserver for the answer */
#ifdef DEBUG
            printf("Querying nameserver ");
            debug_show_ip(nsip);
	    printf("\n");
#endif /* DEBUG */
	    result = query_nameserver(nsip,query,copy);
#ifdef DEBUG
	    printf("%d\n",result);
#endif /* DEBUG */
	    js_destroy(copy);
	    if(result == JS_SUCCESS) /* If a nameserver was found,
	                                then do a lookup again based 
					on the revised cache */
	        recurse_call(id,sock,client,query,recurse_depth + 1,
		             glueless_level);
	    return JS_SUCCESS;
	    }
        read_unlock();
        } while(bobbit_label(copy) > 0);

    /* We could not find a "closer" name server, which is an error
       condition */
    /* Clean up the request */
    js_destroy(copy);
    return JS_SUCCESS;

    cleanup:
        js_destroy(copy);
    cleanup_nojs:
	return JS_ERROR;
    }

/* Routine which prepares the launch of a detached thread which will 
   recursivly look for the DNS name in question 
   Input:  id of query, socket of query, sockaddr struct of query,
           js_string with host name they are looking for
   Output: JS_ERROR on error, JS_SUCCESS on success
*/

int launch_thread(int id, int sock, struct sockaddr client, js_string *query) {
    /* We have to copy the js_string over because the thread needs her
       own copy */
    js_string *copy;
    dnsreq *req;
    pthread_t thread;
    pthread_attr_t attr;

    /* Set up the attributes for the thread.  Make this a detached 
       thread because we don't look at the return value (the thread
       is talking to the client and maybe updating the cache) */
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);

    /* We need to create some thread-safe data structures.  
       We have a special dnsreq data structure that we use because a 
       thread, when initialized, can only be initialized with a single 
       argument. */

    /* Create the req data structure */
    if((req = js_alloc(1,sizeof(dnsreq))) == 0)
        return JS_ERROR;

    /* Create the 'copy' js_string */
    if((copy = js_create(query->unit_count + 3,1)) == 0) {
        js_dealloc(req);
	return JS_ERROR;
	}

    /* Copy over query to 'copy' */
    if(js_copy(query,copy) == JS_ERROR) {
        js_dealloc(req); js_destroy(copy);
	return JS_ERROR;
	}

    /* Fill out the rest of the req structure */
    req->id = id;
    req->sock = sock;
    req->client = client;
    req->query = copy;

    /* Spawn the thread */
#ifdef DEBUG
    recurse_thread(req);
#else
    if(pthread_create(&thread,&attr,(void *)recurse_thread,(void *)req) != 0)
        {
        js_dealloc(req); js_destroy(copy); return JS_ERROR;
	}
#endif /* DEBUG */

    /* Leave...it is now up to the thread to handle the query and free
       memory used */
    return JS_SUCCESS;

    }

/* Add an IP as a server "closer" to the answer for a given
   branch of the DNS tree 
   Input:  js_string of zone (raw UDP format with 2-byte suffix), 
           js_string with ddip, 
	   action to take if a node is already here:
	   1) Overwrite 2) Append 3) "Mask" currect table
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_jsddip(js_string *zone, js_string *ddip, int if_exists) {
    closer *close;
    js_string *ipjs;
    uint32 ip, *ipp;
    closer *point, *point_save;
    mhash_e spot_data;

    /* Allocate the memory for a new spot on the dnscache hash */
    if((close = js_alloc(1,sizeof(closer))) == 0)
        return JS_ERROR;

    /* Allocate the memory for the IP we will put on the hash */
    if((ipjs = js_create(6,1)) == 0) {
        js_dealloc(close);
	return JS_ERROR;
	}

    /* ddip_2_ip uses js_atoi, which means we have to set the encoding */
    ddip->encoding = MARA_LOCALE;

    /* Convert their ddip in to an ip */
    if(ddip_2_ip(ddip,ipjs,0) == JS_ERROR)
        goto cleanup;

    /* Make sure the raw zone is of the right rtype */
    if(change_rtype(zone,252) == JS_ERROR)
        goto cleanup;

    /* Convert ipjs in to a format that sendto() can use */
    if(ipjs->unit_count < 4)
        goto cleanup;
    ip = (*(ipjs->string) << 24) | (*(ipjs->string + 1) << 16) | 
         (*(ipjs->string + 2) << 8) | *(ipjs->string + 3);

    /* Fill up the "close" structure */
    close->num_elements = 1;
    close->datatype = RR_A;
    if((ipp = js_alloc(1,sizeof(uint32))) == 0)
        goto cleanup;
    *ipp = ip;
    close->data = ipp;
    close->next = NULL;

    /* Add the data to the hash */
    write_lock();
    /* See if the data is already there.  If so, act based on the value
       of if_exists */
    spot_data = mhash_get(dnscache,zone);
    if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NS) {
        point = spot_data.value;
	switch(if_exists) {
	    case 1: /* Overwrite */
	        /* Destroy the pointer to the element in the hash */
		mhash_undef(dnscache,zone);
		/* Destroy the element itself.  Go down the
		   linked list ans destroy each element one by one,
		   starting at the top. */
	        while(point != 0) {
	           point_save = point->next;
		   if(point->datatype == RR_A && point->data != 0) 
		       js_dealloc(point->data);
                   else if(point->datatype == RR_NS && point->data != 0)
		       js_destroy(point->data);
                   else if(point->data != 0)
		       js_dealloc(point->data);
                   js_dealloc(point);
		   point = point_save;
                   }
                break; 
	    case 2: /* Append */
                while(point->next != 0) 
	            point = point->next; 
                point->next = close;
		return JS_SUCCESS;
            /* XXX to do: mask */
	    }
        }
    /* This error condition should never happen */
    else if(spot_data.value != 0)
        return JS_ERROR;
    /* It's a (possibly newly) empty spot--add a new entry to the hash 
       (w/ error checking) */
    else if(mhash_put(dnscache,zone,close,MARA_DNS_NS) == JS_ERROR) {
        write_unlock();
	goto cleanup;
	}
    write_unlock();

    return JS_SUCCESS;
    cleanup_ipp:
        js_dealloc(ipp);
    cleanup:
        js_destroy(ipjs);
	js_dealloc(close);
	return JS_ERROR;

    }

/* Add the name of a server "closer" to the answer for a given
   branch of the DNS tree 
   Input:  js_string of zone (raw UDP format with 2-byte suffix), 
           js_string with server name
	   action to take if a node is already here:
	   1) Overwrite 2) Append 3) "Mask" currect table
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_js(js_string *zone, js_string *name, int if_exists) {
    closer *close;
    mhash_e spot_data;
    closer *point, *point_save;
    js_string *new;

    /* Allocate the memory for a new spot on the dnscache hash */
    if((close = js_alloc(1,sizeof(closer))) == 0)
        return JS_ERROR;
    if((new = js_create(name->unit_count + 1,1)) == 0)
        return JS_ERROR;

    if(js_copy(name,new) == JS_ERROR)
        return JS_ERROR;

    /* Make sure the raw zone is of the right rtype */
    if(change_rtype(zone,252) == JS_ERROR)
        goto cleanup;

    /* Fill up the "close" structure */
    close->num_elements = 1;
    close->datatype = RR_NS;
    close->data = new;
    close->next = NULL;

    /* Add the data to the hash */
    write_lock();
    /* See if the data is already there.  If so, add this to the end of
       the chain */
    spot_data = mhash_get(dnscache,zone);

    if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NS) {
        point = spot_data.value;
	switch(if_exists) {
	    case 1: /* Overwrite */
	        /* Destroy the pointer to the element in the hash */
		mhash_undef(dnscache,zone);
		/* Destroy the element itself.  Go down the
		   linked list ans destroy each element one by one,
		   starting at the top. */
	        while(point != 0) {
	           point_save = point->next;
		   if(point->datatype == RR_A && point->data != 0) 
		       js_dealloc(point->data);
                   else if(point->datatype == RR_NS && point->data != 0)
		       js_destroy(point->data);
                   else if(point->data != 0)
		       js_dealloc(point->data);
                   js_dealloc(point);
		   point = point_save;
                   }
                break; 
	    case 2: /* Append */
                while(point->next != 0) 
	            point = point->next; 
                point->next = close;
		return JS_SUCCESS;
            /* XXX to do: mask */
	    }
        }
    /* This error condition should never happen */
    else if(spot_data.value != 0)
        return JS_ERROR;

    /* It's an empty spot--add a new entry to the hash (w/ error checking) */
    else if(mhash_put(dnscache,zone,close,MARA_DNS_NS) == JS_ERROR) {
        write_unlock();
	goto cleanup;
	}
    write_unlock();

    return JS_SUCCESS;
    cleanup:
        js_destroy(new);
	return JS_ERROR;

    }

/* Wrapper that allows a substring of another js_string object to be
   the zone we add a "closer" jsip for.
   Input: js_string object with desired dlabel as substring
          offset of where the dlabel begins
	  ipjs of IP to add
	  action to do if node is already there
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_jsip_offset(js_string *js, int offset, js_string *ipjs,
                           int if_exists) {

    js_string *sub;
    int len, ret;

    /* Determine how long the domain label is */
    len = dlabel_length(js,offset);
    if(len == JS_ERROR)
        return JS_ERROR;

    /* Make a string which holds just the domain name label */
    if((sub = js_create(len + 4,1)) == 0)
        return JS_ERROR;
    if(js_substr(js,sub,offset,len) == JS_ERROR) {
        js_destroy(sub);
	return JS_ERROR;
	}
    if(js_adduint16(sub,252) == JS_ERROR) {
        js_destroy(sub);
	return JS_ERROR;
	}
     

    /* Run the actual add_closer_jsip function */
    ret = add_closer_jsip(sub,ipjs,if_exists);
    js_destroy(sub);
    return ret;
    }

/* Add an IP as a server "closer" to the answer for a given
   branch of the DNS tree 
   Input:  js_string of zone (raw UDP format with 2-byte suffix), 
           js_string with ip
	   action to take if a node is already here:
	   1) Overwrite 2) Append 3) "Mask" currect table
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_jsip(js_string *zone, js_string *ipjs, int if_exists) {
    closer *close;
    uint32 ip, *ipp;
    closer *point, *point_save;
    mhash_e spot_data;

    /* Allocate the memory for a new spot on the dnscache hash */
    if((close = js_alloc(1,sizeof(closer))) == 0)
        return JS_ERROR;

    /* Make sure the raw zone is of the right rtype */
    if(change_rtype(zone,252) == JS_ERROR)
        goto cleanup;

    /* Convert ipjs in to a format that sendto() can use */
    if(ipjs->unit_count < 4)
        goto cleanup;
    ip = (*(ipjs->string) << 24) | (*(ipjs->string + 1) << 16) | 
         (*(ipjs->string + 2) << 8) | *(ipjs->string + 3);

    /* Fill up the "close" structure */
    close->num_elements = 1;
    close->datatype = RR_A;
    if((ipp = js_alloc(1,sizeof(uint32))) == 0)
        goto cleanup;
    *ipp = ip;
    close->data = ipp;
    close->next = NULL;

    /* Add the data to the hash */
    write_lock();
    /* See if the data is already there.  If so, add this to the end of
       the chain */
    spot_data = mhash_get(dnscache,zone);

    if(spot_data.value != 0 && spot_data.datatype == MARA_DNS_NS) {
        point = spot_data.value;
	switch(if_exists) {
	    case 1: /* Overwrite */
	        /* Destroy the pointer to the element in the hash */
		mhash_undef(dnscache,zone);
		/* Destroy the element itself.  Go down the
		   linked list ans destroy each element one by one,
		   starting at the top. */
	        while(point != 0) {
	           point_save = point->next;
		   if(point->datatype == RR_A && point->data != 0) 
		       js_dealloc(point->data);
                   else if(point->datatype == RR_NS && point->data != 0)
		       js_destroy(point->data);
                   else if(point->data != 0)
		       js_dealloc(point->data);
                   js_dealloc(point);
		   point = point_save;
                   }
                break; 
	    case 2: /* Append */
                while(point->next != 0) 
	            point = point->next; 
                point->next = close;
		return JS_SUCCESS;
            /* XXX to do: mask */
	    }
        }
    /* This error condition should never happen */
    else if(spot_data.value != 0)
        return JS_ERROR;

    /* It's an empty spot--add a new entry to the hash (w/ error checking) */
    else if(mhash_put(dnscache,zone,close,MARA_DNS_NS) == JS_ERROR) {
        write_unlock();
	goto cleanup;
	}
    write_unlock();

    return JS_SUCCESS;
    cleanup_ipp:
        js_dealloc(ipp);
    cleanup:
        js_destroy(ipjs);
	js_dealloc(close);
	return JS_ERROR;

    }

/* Add an IP as a server "closer" to the answer for a given
   branch of the DNS tree 
   Input:  null-terminated string of zone, nt-string of ddip
   Output: JS_ERROR on fatal error, JS_SUCCESS on success
*/

int add_closer_ip(char *zone, char *ddip) {
    js_string *zone_js, *ddip_js;
    int zone_len, ddip_len, ret;

    /* Sanity checks */
    if(zone == 0 || *zone == 0)
        return JS_ERROR;
    if(ddip == 0 || *ddip == 0)
        return JS_ERROR;

    zone_len = strlen(zone);
    ddip_len = strlen(ddip);

    /* Initialize the js string objects */
    if((zone_js = js_create(zone_len + 3,1)) == 0)
        return JS_ERROR;
 
    if((ddip_js = js_create(ddip_len + 1,1)) == 0) {
        js_destroy(zone_js);
        return JS_ERROR; 
	}

    /* Copy over the null-terminated strings to the js string objects */
    if(js_str2js(zone_js,zone,zone_len,1) == JS_ERROR)
        goto cleanup;
    if(js_str2js(ddip_js,ddip,ddip_len,1) == JS_ERROR)
        goto cleanup;

    /* Convert the zone name in to a binary string */
    if(hname_2rfc1035(zone_js) <= 0)
        goto cleanup;
    /* Add a two-byte "query type" to the end of "zone_js" */
    if(js_adduint16(zone_js,252) == JS_ERROR)
        goto cleanup;
 
    /* Call the "real" function that uses the js_string objects */
    ret = add_closer_jsddip(zone_js,ddip_js,OVERWRITE);
    js_destroy(zone_js);
    js_destroy(ddip_js);
    return ret;

    cleanup:
        js_destroy(ddip_js);
        js_destroy(zone_js);
        return JS_ERROR; 
    }

/* Initialize the cache hash for general use
   Input: none (parameters obtained from kvar variables)
   Output: JS_ERROR on fatal error, JS_SUCCESS on success 
*/

int init_cache() { 
    closer root;
    js_string rootns;
    js_string rootns_ip;

    if(dnscache != 0) /* We can only init the DNS cache once */
        return JS_ERROR;

    /* Initialize the dnscache object */
    dnscache = (mhash *)mhash_create(12); /* 1024+ usable elements, 4096
                                             total hash table spots */

    add_closer_ip("A.","127.0.0.1");

    return JS_SUCCESS;
    }
